# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for structured parsing."""

import asyncio
import inspect
import unittest

import langfun.core as lf
from langfun.core.llms import fake
from langfun.core.structured import mapping
from langfun.core.structured import parsing
import pyglove as pg


class Activity(pg.Object):
  description: str


class Itinerary(pg.Object):
  day: pg.typing.Int[1, None]
  type: pg.typing.Enum['daytime', 'nighttime']
  activities: list[Activity]
  hotel: pg.typing.Str['.*Hotel'] | None


class ParseStructurePythonTest(unittest.TestCase):

  def test_render_no_examples(self):
    l = parsing._ParseStructurePython(int)
    m = lf.AIMessage('12 / 6 + 2 = 4')
    self.assertEqual(
        l.render(input=m, context='Compute 12 / 6 + 2.').text,
        inspect.cleandoc("""
            Please help translate the last LM_RESPONSE into RESULT_OBJECT based on RESULT_TYPE.

            INSTRUCTIONS:
            1. Both RESULT_TYPE and RESULT_OBJECT are described in Python.
            2. RESULT_OBJECT must be created solely based on the information provided in LM_RESPONSE.

            USER_REQUEST:
              Compute 12 / 6 + 2.

            LM_RESPONSE:
              12 / 6 + 2 = 4

            RESULT_TYPE:
              int

            RESULT_OBJECT:
            """),
    )

  def test_render_no_context(self):
    l = parsing._ParseStructurePython(int)
    m = lf.AIMessage('12 / 6 + 2 = 4')

    self.assertEqual(
        l.render(input=m).text,
        inspect.cleandoc("""
            Please help translate the last LM_RESPONSE into RESULT_OBJECT based on RESULT_TYPE.

            INSTRUCTIONS:
            1. Both RESULT_TYPE and RESULT_OBJECT are described in Python.
            2. RESULT_OBJECT must be created solely based on the information provided in LM_RESPONSE.

            LM_RESPONSE:
              12 / 6 + 2 = 4

            RESULT_TYPE:
              int

            RESULT_OBJECT:
            """),
    )

  def test_render(self):
    l = parsing._ParseStructurePython(
        int,
        examples=[
            mapping.MappingExample(
                context='What is the answer of 1 plus 1?',
                input='1 + 1 = 2',
                output=2,
            ),
            mapping.MappingExample(
                context='Compute the value of 3 + (2 * 6).',
                input='fifteen',
                output=15,
            ),
        ],
    )
    self.assertEqual(
        l.render(input=lf.AIMessage('Compute 12 / 6 + 2.')).text,
        inspect.cleandoc("""
            Please help translate the last LM_RESPONSE into RESULT_OBJECT based on RESULT_TYPE.

            INSTRUCTIONS:
            1. Both RESULT_TYPE and RESULT_OBJECT are described in Python.
            2. RESULT_OBJECT must be created solely based on the information provided in LM_RESPONSE.

            USER_REQUEST:
              What is the answer of 1 plus 1?

            LM_RESPONSE:
              1 + 1 = 2

            RESULT_TYPE:
              int

            RESULT_OBJECT:
              ```python
              2
              ```

            USER_REQUEST:
              Compute the value of 3 + (2 * 6).

            LM_RESPONSE:
              fifteen

            RESULT_TYPE:
              int

            RESULT_OBJECT:
              ```python
              15
              ```


            LM_RESPONSE:
              Compute 12 / 6 + 2.

            RESULT_TYPE:
              int

            RESULT_OBJECT:
            """),
    )

  def test_invocation(self):
    lm_input = '3-day itineraries to San Francisco'
    lm_response = inspect.cleandoc("""
        **Day 1:**

        * Arrive in San Francisco and check into your hotel.
        * Take a walk around Fisherman's Wharf and have dinner at one of the
          many seafood restaurants.
        * Visit Pier 39 and see the sea lions.

        **Day 2:**

        * Take a ferry to Alcatraz Island and tour the infamous prison.
        * Take a walk across the Golden Gate Bridge.
        * Visit the Japanese Tea Garden in Golden Gate Park.

        **Day 3:**

        * Visit the de Young Museum and see the collection of American art.
        * Visit the San Francisco Museum of Modern Art.
        * Take a cable car ride.
        """)

    parse_structured_response = inspect.cleandoc(
        """
        ```python
        [
            Itinerary(
                day=1,
                type='daytime',
                activities=[
                    Activity(description='Arrive in San Francisco and check into your hotel.'),
                    Activity(description='Take a walk around Fisherman\\'s Wharf and have dinner at one of the many seafood restaurants.'),
                    Activity(description='Visit Pier 39 and see the sea lions.'),
                ], 
                hotel=None),
            Itinerary(
                day=2,
                type='daytime',
                activities=[
                    Activity(description='Take a ferry to Alcatraz Island and tour the infamous prison.'),
                    Activity(description='Take a walk across the Golden Gate Bridge.'),
                    Activity(description='Visit the Japanese Tea Garden in Golden Gate Park.'),
                ], 
                hotel=None),
            Itinerary(
                day=3,
                type='daytime',
                activities=[
                    Activity(description='Visit the de Young Museum and see the collection of American art.'),
                    Activity(description='Visit the San Francisco Museum of Modern Art.'),
                    Activity(description='Take a cable car ride.'),
                ], 
                hotel=None),
        ]
        ```
        """)

    with lf.context(
        lm=fake.StaticSequence(
            [lm_response, parse_structured_response],
        ),
        override_attrs=True,
    ):
      l = parsing._ParseStructurePython(
          [Itinerary],
          examples=[
              mapping.MappingExample(
                  context=inspect.cleandoc("""
                      Find the alternatives of expressing \"feeling great\".
                      """),
                  input=inspect.cleandoc("""
                      Here are some words for expressing \"feeling great\".
                      * Ecstatic
                      * Delighted
                      * Wonderful
                      * Enjoyable
                      * Fantastic"""),
                  schema={'expression': str, 'words': list[str]},
                  output={
                      'expression': 'feeling great',
                      'words': [
                          'Ecstatic',
                          'Delighted',
                          'Wonderful',
                          'Enjoyable',
                          'Fantastic',
                      ],
                  },
              )
          ],
      )
      m = lf.LangFunc(lm_input)()
      r = l(input=m)
      self.assertEqual(len(r.result), 3)
      self.assertIsInstance(r.result[0], Itinerary)
      self.assertEqual(len(r.result[0].activities), 3)
      self.assertIsNone(r.result[0].hotel)

  def test_bad_response(self):
    with lf.context(
        lm=fake.StaticResponse('a3'),
        override_attrs=True,
    ):
      with self.assertRaisesRegex(
          mapping.MappingError,
          'name .* is not defined',
      ):
        parsing.parse('three', int)

  def test_autofix(self):
    lm = fake.StaticSequence([
        '=3',
        inspect.cleandoc("""
            CorrectedCode(
                corrected_code='3',
            )
            """),
    ])
    self.assertEqual(parsing.parse('three', int, lm=lm, autofix=3), 3)

  def test_parse(self):
    lm = fake.StaticResponse('1')
    self.assertEqual(parsing.parse('the answer is 1', int, lm=lm), 1)
    self.assertEqual(
        parsing.parse(
            'the answer is 1', int, user_prompt='what is 0 + 1?', lm=lm
        ),
        1,
    )
    r = parsing.parse(
        'the answer is 1', int, user_prompt='what is 0 + 1?', lm=lm,
        returns_message=True
    )
    self.assertEqual(
        r,
        lf.AIMessage(
            '1', score=1.0, result=1, logprobs=None, is_cached=False,
            usage=lf.LMSamplingUsage(652, 1, 653),
            tags=['lm-response', 'lm-output', 'transformed']
        ),
    )

  def test_aparse(self):
    with lf.context(lm=fake.StaticResponse('1')):
      self.assertEqual(asyncio.run(parsing.aparse('the answer is 1', int)), 1)


class ParseStructureJsonTest(unittest.TestCase):

  def test_render_no_examples(self):
    l = parsing._ParseStructureJson(int)
    m = lf.AIMessage('12 / 6 + 2 = 4')
    self.assertEqual(
        l.render(input=m, context='Compute 12 / 6 + 2.').text,
        inspect.cleandoc("""
            Please help translate the last LM response into JSON based on the request and the schema:

            INSTRUCTIONS:
              1. If the schema has `_type`, carry it over to the JSON output.
              2. If a field from the schema cannot be extracted from the response, use null as the JSON value.

            USER_REQUEST:
              Compute 12 / 6 + 2.

            LM_RESPONSE:
              12 / 6 + 2 = 4

            SCHEMA:
              {"result": int}

            JSON:
            """),
    )

  def test_render_no_context(self):
    l = parsing._ParseStructureJson(int)
    m = lf.AIMessage('12 / 6 + 2 = 4')

    self.assertEqual(
        l.render(input=m).text,
        inspect.cleandoc("""
            Please help translate the last LM response into JSON based on the request and the schema:

            INSTRUCTIONS:
              1. If the schema has `_type`, carry it over to the JSON output.
              2. If a field from the schema cannot be extracted from the response, use null as the JSON value.

            LM_RESPONSE:
              12 / 6 + 2 = 4

            SCHEMA:
              {"result": int}

            JSON:
            """),
    )

  def test_render(self):
    l = parsing._ParseStructureJson(
        int,
        examples=[
            mapping.MappingExample(
                context='What is the answer of 1 plus 1?',
                input='1 + 1 = 2',
                output=2,
            ),
            mapping.MappingExample(
                context='Compute the value of 3 + (2 * 6).',
                input='fifteen',
                output=15,
            ),
        ],
    )
    self.assertEqual(
        l.render(input=lf.AIMessage('Compute 12 / 6 + 2.')).text,
        inspect.cleandoc("""
            Please help translate the last LM response into JSON based on the request and the schema:

            INSTRUCTIONS:
              1. If the schema has `_type`, carry it over to the JSON output.
              2. If a field from the schema cannot be extracted from the response, use null as the JSON value.

            USER_REQUEST:
              What is the answer of 1 plus 1?

            LM_RESPONSE:
              1 + 1 = 2

            SCHEMA:
              {"result": int}

            JSON:
              {"result": 2}

            USER_REQUEST:
              Compute the value of 3 + (2 * 6).

            LM_RESPONSE:
              fifteen

            SCHEMA:
              {"result": int}

            JSON:
              {"result": 15}


            LM_RESPONSE:
              Compute 12 / 6 + 2.

            SCHEMA:
              {"result": int}

            JSON:
            """),
    )

  def test_invocation(self):
    lm_input = '3-day itineraries to San Francisco'
    lm_response = inspect.cleandoc("""
        **Day 1:**

        * Arrive in San Francisco and check into your hotel.
        * Take a walk around Fisherman's Wharf and have dinner at one of the
          many seafood restaurants.
        * Visit Pier 39 and see the sea lions.

        **Day 2:**

        * Take a ferry to Alcatraz Island and tour the infamous prison.
        * Take a walk across the Golden Gate Bridge.
        * Visit the Japanese Tea Garden in Golden Gate Park.

        **Day 3:**

        * Visit the de Young Museum and see the collection of American art.
        * Visit the San Francisco Museum of Modern Art.
        * Take a cable car ride.
        """)

    parse_structured_response = (
        lf.LangFunc(
            """
        {"result": [
          {
            "_type": {{itinerary_type}},
            "day": 1,
            "type": "daytime",
            "activities": [
              {
                "_type": {{activity_type}},
                "description": "Arrive in San Francisco and check into your hotel."
              },
              {
                "_type": {{activity_type}},
                "description": "Take a walk around Fisherman's Wharf and have dinner at one of the many seafood restaurants."
              },
              {
                "_type": {{activity_type}},
                "description": "Visit Pier 39 and see the sea lions."
              }
            ],
            "hotel": null
          },
          {
              "_type": {{itinerary_type}},
              "day": 2,
              "type": "daytime",
              "activities": [
                {
                  "_type": {{activity_type}},
                  "description": "Take a ferry to Alcatraz Island and tour the infamous prison."
                },
                {
                  "_type": {{activity_type}},
                  "description": "Take a walk across the Golden Gate Bridge."
                },
                {
                  "_type": {{activity_type}},
                  "description": "Visit the Japanese Tea Garden in Golden Gate Park."
                }
              ], 
              "hotel": null
           },
           {
              "_type": {{itinerary_type}},
              "day": 3,
              "type": "daytime",
              "activities": [
                {
                  "_type": {{activity_type}},
                  "description": "Visit the de Young Museum and see the collection of American art."
                },
                {
                  "_type": {{activity_type}},
                  "description": "Visit the San Francisco Museum of Modern Art."
                },
                {
                  "_type": {{activity_type}},
                  "description": "Take a cable car ride."
                }
              ],
              "hotel": null
            }
          ]}
        """,
            itinerary_type=f'"{Itinerary.__type_name__}"',
            activity_type=f'"{Activity.__type_name__}"',
        )
        .render()
        .text
    )
    with lf.context(
        lm=fake.StaticSequence(
            [lm_response, parse_structured_response],
        ),
        override_attrs=True,
    ):
      message = lf.LangFunc(lm_input)()
      l = parsing._ParseStructureJson(
          [Itinerary],
          examples=[
              mapping.MappingExample(
                  context=inspect.cleandoc("""
                      Find the alternatives of expressing \"feeling great\".
                      """),
                  input=inspect.cleandoc("""
                      Here are some words for expressing \"feeling great\".
                      * Ecstatic
                      * Delighted
                      * Wonderful
                      * Enjoyable
                      * Fantastic"""),
                  schema={'expression': str, 'words': list[str]},
                  output={
                      'expression': 'feeling great',
                      'words': [
                          'Ecstatic',
                          'Delighted',
                          'Wonderful',
                          'Enjoyable',
                          'Fantastic',
                      ],
                  },
              )
          ],
      )
      r = l(input=message)
      self.assertEqual(len(r.result), 3)
      self.assertIsInstance(r.result[0], Itinerary)
      self.assertEqual(len(r.result[0].activities), 3)
      self.assertIsNone(r.result[0].hotel)

  def test_bad_response(self):
    with lf.context(
        lm=fake.StaticSequence(['`3`']),
        override_attrs=True,
    ):
      with self.assertRaisesRegex(
          mapping.MappingError,
          'No JSON dict in the output',
      ):
        parsing.parse('three', int, protocol='json')

    with lf.context(
        lm=fake.StaticSequence(['`3`']),
        override_attrs=True,
    ):
      # Test default.
      self.assertIsNone(parsing.parse('three', int, default=None))

  def test_parse(self):
    lm = fake.StaticResponse('{"result": 1}')
    self.assertEqual(
        parsing.parse('the answer is 1', int, lm=lm, protocol='json'), 1
    )
    self.assertEqual(
        parsing.parse(
            'the answer is 1',
            int,
            user_prompt='what is 0 + 1?',
            lm=lm,
            protocol='json',
        ),
        1,
    )


class CallTest(unittest.TestCase):

  def test_call_with_const_str(self):
    with lf.context(
        lm=fake.StaticMapping({
            'Compute 1 + 2': 'three',
        })
    ):
      self.assertEqual(parsing.call('Compute 1 + 2'), 'three')

  def test_acall(self):
    with lf.context(
        lm=fake.StaticMapping({
            'Compute 1 + 2': 'three',
        })
    ):
      self.assertEqual(
          asyncio.run(parsing.acall('Compute 1 + 2')), 'three'
      )

  def test_call_with_template_str(self):
    with lf.context(
        lm=fake.StaticMapping({
            'Compute 1 + 2': 'three',
        })
    ):
      self.assertEqual(parsing.call('Compute {{x}} + {{y}}', x=1, y=2), 'three')

  def test_call_with_explicit_template(self):
    with lf.context(
        lm=fake.StaticMapping({
            'Compute 1 + 2': 'three',
        })
    ):
      self.assertEqual(
          parsing.call(lf.Template('Compute {{x}} + {{y}}', x=1, y=2)), 'three'
      )

    with lf.context(
        lm=fake.StaticMapping({
            'Compute 1 + 2': 'three',
        })
    ):
      self.assertEqual(
          parsing.call(lf.Template('Compute {{x}} + {{y}}'), x=1, y=2), 'three'
      )

  def test_call_with_lfun(self):
    with lf.context(
        lm=fake.StaticMapping({
            'Compute 1 + 2': 'three',
        })
    ):
      self.assertEqual(
          parsing.call(lf.LangFunc('Compute {{x}} + {{y}}', x=1, y=2)), 'three'
      )

  def test_call_with_schema(self):
    self.assertEqual(
        parsing.call(
            'Compute 1 + 2', int, lm=fake.StaticSequence(['three', '3'])
        ),
        3,
    )

    with lf.context(lm=fake.StaticSequence(['three', '3'])):
      self.assertEqual(
          parsing.call(lf.LangFunc('Compute {{x}} + {{y}}', x=1, y=2), int), 3
      )

  def test_call_with_returning_message(self):
    r = parsing.call(
        'Compute 1 + 2', int, lm=fake.StaticSequence(['three', '3']),
        returns_message=True
    )
    self.assertEqual(
        r,
        lf.AIMessage(
            '3',
            result=3,
            score=1.0,
            logprobs=None,
            is_cached=False,
            usage=lf.LMSamplingUsage(356, 1, 357),
            tags=['lm-response', 'lm-output', 'transformed']
        ),
    )

  def test_call_with_parsing_args(self):
    self.assertEqual(
        parsing.call(
            'Compute 1 + 2',
            int,
            lm=fake.StaticSequence(['three']),
            parsing_lm=fake.StaticSequence(['3']),
            parsing_examples=[
                mapping.MappingExample(
                    context='Multiple four and five',
                    input='twenty',
                    schema=int,
                    output=20,
                )
            ],
        ),
        3,
    )

  def test_call_with_parsing_message_chaining(self):
    output = parsing.call(
        'Compute 1 + 2',
        int,
        lm=fake.StaticSequence(['three']),
        parsing_lm=fake.StaticSequence(['3']),
        parsing_examples=[
            mapping.MappingExample(
                context='Multiple four and five',
                input='twenty',
                schema=int,
                output=20,
            )
        ],
        returns_message=True,
    )
    self.assertIn('parsing-lm-output', output.tags)
    self.assertIn('parsing-lm-input', output.source.tags)
    self.assertEqual(output.root.text, 'Compute 1 + 2')

  def test_call_with_parsing_message_chaining_on_parsing_error(self):
    try:
      output = parsing.call(
          'Compute 1 + 2',
          int,
          lm=fake.StaticSequence(['three']),
          parsing_lm=fake.StaticSequence(['abc']),
          parsing_examples=[
              mapping.MappingExample(
                  context='Multiple four and five',
                  input='twenty',
                  schema=int,
                  output=20,
              )
          ],
          returns_message=True,
      )
    except mapping.MappingError as e:
      output = e.lm_response
    self.assertIn('parsing-lm-output', output.tags)
    self.assertIn('parsing-lm-input', output.source.tags)
    self.assertEqual(output.root.text, 'Compute 1 + 2')

  def test_call_with_autofix(self):
    lm = fake.StaticSequence(
        [
            'three',
            '=3',
            inspect.cleandoc("""
            CorrectedCode(
                corrected_code='3',
            )
            """),
        ],
        debug=True,
    )
    self.assertEqual(
        parsing.call('what is one plus two?', int, lm=lm, autofix=3), 3
    )

  def test_call_with_structured_input(self):
    self.assertEqual(parsing.call(1, lm=fake.StaticResponse('2')), '2')

  def test_call_with_response_postprocess(self):
    target_str = '@TARGET_STR@'
    random_str = '!RANDOM_STR!'
    delimiter = '\n'

    raw_response = target_str + delimiter + random_str
    r = parsing.call(
        'Compute 1 + 2',
        int,
        lm=fake.StaticSequence([raw_response, '3']),
        returns_message=True,
        response_postprocess=lambda x: x.split(delimiter)[0],
    )
    self.assertIn(target_str, str(r.lm_input))
    self.assertNotIn(random_str, str(r.lm_input))


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""MIME tests."""
import inspect
import unittest
from unittest import mock

import langfun.core as lf
from langfun.core.modalities import mime
import pyglove as pg


def mock_request(*args, **kwargs):
  del args, kwargs
  return pg.Dict(content=b'foo')


def mock_readfile(*args, **kwargs):
  del args, kwargs
  return b'bar'


class CustomMimeTest(unittest.TestCase):

  def test_is_text(self):
    self.assertTrue(mime.Custom('text/plain', b'foo').is_text)
    self.assertTrue(mime.Custom('text/xml', b'foo').is_text)
    self.assertTrue(mime.Custom('application/json', b'foo').is_text)
    self.assertTrue(mime.Custom('application/x-python-code', b'foo').is_text)
    self.assertFalse(mime.Custom('application/pdf', b'foo').is_text)
    self.assertFalse(mime.Custom('application/octet-stream', b'foo').is_text)

  def test_from_byes(self):
    content = mime.Mime.from_bytes(b'hello')
    self.assertIs(content.__class__, mime.Mime)

    content = mime.Custom('text/plain', b'foo')
    self.assertEqual(content.to_bytes(), b'foo')
    self.assertEqual(content.mime_type, 'text/plain')
    self.assertTrue(content.is_text)
    self.assertFalse(content.is_binary)
    self.assertEqual(content.to_text(), 'foo')
    self.assertTrue(content.is_compatible('text/plain'))
    self.assertFalse(content.is_compatible('text/xml'))
    self.assertIs(content.make_compatible('text/plain'), content)

    with self.assertRaisesRegex(
        lf.ModalityError, '.* cannot be converted to supported types'
    ):
      content.make_compatible('application/pdf')

    with self.assertRaisesRegex(
        ValueError, 'Either uri or content must be provided.'
    ):
      mime.Custom('text/plain')

  def test_uri(self):
    content = mime.Custom.from_uri('http://mock/web/a.txt', mime='text/plain')
    with mock.patch('requests.get') as mock_requests_stub:
      mock_requests_stub.side_effect = mock_request
      self.assertEqual(content.uri, 'http://mock/web/a.txt')
      self.assertEqual(content.content_uri, 'data:text/plain;base64,Zm9v')
      self.assertEqual(content.embeddable_uri, 'http://mock/web/a.txt')

    content = mime.Custom.from_uri('a.txt', mime='text/plain')
    with mock.patch('pyglove.io.readfile') as mock_readfile_stub:
      mock_readfile_stub.side_effect = mock_readfile
      self.assertEqual(content.uri, 'a.txt')
      self.assertEqual(content.content_uri, 'data:text/plain;base64,YmFy')
      self.assertEqual(content.embeddable_uri, 'data:text/plain;base64,YmFy')

  def test_from_uri(self):
    content = mime.Custom.from_uri('http://mock/web/a.txt', mime='text/plain')
    with mock.patch('requests.get') as mock_requests_stub:
      mock_requests_stub.side_effect = mock_request
      self.assertEqual(content.to_bytes(), b'foo')
      self.assertEqual(content.mime_type, 'text/plain')

    content = mime.Custom.from_uri('a.txt', mime='text/plain')
    with mock.patch('pyglove.io.readfile') as mock_readfile_stub:
      mock_readfile_stub.side_effect = mock_readfile
      self.assertEqual(content.to_bytes(), b'bar')
      self.assertEqual(content.mime_type, 'text/plain')

    content = mime.Mime.from_uri('data:text/plain;base64,Zm9v')
    self.assertIsNone(content.uri)
    self.assertEqual(content.mime_type, 'text/plain')
    self.assertEqual(content.content, b'foo')
    self.assertEqual(content.content_uri, 'data:text/plain;base64,Zm9v')
    self.assertEqual(content.embeddable_uri, 'data:text/plain;base64,Zm9v')

    with self.assertRaisesRegex(ValueError, 'Invalid data URI'):
      mime.Mime.from_uri('data:text/plain')

    with self.assertRaisesRegex(ValueError, 'Invalid data URI'):
      mime.Mime.from_uri('data:text/plain;abcd')

    with self.assertRaisesRegex(ValueError, 'Unsupported encoding'):
      mime.Mime.from_uri('data:text/plain;base16,abcd')

  def assert_html_content(self, html, expected):
    expected = inspect.cleandoc(expected).strip()
    actual = html.content.strip()
    if actual != expected:
      print(actual)
    self.assertEqual(actual, expected)

  def test_html(self):
    self.assert_html_content(
        mime.Custom('text/plain', b'foo').to_html(
            enable_summary_tooltip=False,
            enable_key_tooltip=False,
        ),
        """
        <details open class="pyglove custom"><summary><div class="summary-title">Custom(...)</div></summary><embed type="text/plain" src="data:text/plain;base64,Zm9v"/></details>
        """
    )
    self.assert_html_content(
        mime.Custom('text/plain', b'foo').to_html(
            enable_summary_tooltip=False,
            enable_key_tooltip=False,
            extra_flags=dict(
                raw_mime_content=True,
            )
        ),
        """
        <embed type="text/plain" src="data:text/plain;base64,Zm9v"/>
        """
    )
    self.assert_html_content(
        mime.Custom('text/plain', b'foo').to_html(
            enable_summary_tooltip=False,
            enable_key_tooltip=False,
            extra_flags=dict(
                display_modality_when_hover=True,
            )
        ),
        """
        <details open class="pyglove custom"><summary><div class="summary-title">Custom(...)</div><span class="tooltip"><embed type="text/plain" src="data:text/plain;base64,Zm9v"/></span></summary><embed type="text/plain" src="data:text/plain;base64,Zm9v"/></details>
        """
    )


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest
import langfun.core as lf
from langfun.core.llms import deepseek


class DeepSeekTest(unittest.TestCase):
  """Tests for DeepSeek language model."""

  def test_dir(self):
    self.assertIn('deepseek-v3', deepseek.DeepSeek.dir())

  def test_key(self):
    with self.assertRaisesRegex(ValueError, 'Please specify `api_key`'):
      _ = deepseek.DeepSeekV3().headers
    self.assertEqual(
        deepseek.DeepSeekV3(api_key='test_key').headers,
        {
            'Content-Type': 'application/json',
            'Authorization': 'Bearer test_key',
        }
    )

  def test_model_id(self):
    self.assertEqual(
        deepseek.DeepSeekV3(api_key='test_key').model_id,
        'deepseek-v3',
    )

  def test_resource_id(self):
    self.assertEqual(
        deepseek.DeepSeekV3(api_key='test_key').resource_id,
        'deepseek://deepseek-v3',
    )

  def test_request(self):
    request = deepseek.DeepSeekV3(api_key='test_key').request(
        lf.UserMessage('hi'), lf.LMSamplingOptions(temperature=0.0),
    )
    self.assertEqual(request['model'], 'deepseek-chat')

  def test_lm_get(self):
    self.assertIsInstance(
        lf.LanguageModel.get('deepseek-v3'), deepseek.DeepSeek
    )

if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""langfun features on Python code parsing and execution."""

# pylint: disable=g-bad-import-order
# pylint: disable=g-importing-member

# Expose from `lf.coding` as aliases for `pg.coding` for backward compatibility.
from langfun.core.coding.python.execution import CodeError
from langfun.core.coding.python.execution import CodePermission
from langfun.core.coding.python.execution import context

from langfun.core.coding.python.parsing import clean
from langfun.core.coding.python.execution import evaluate
from langfun.core.coding.python.execution import run

from langfun.core.coding.python.generation import PythonCode
from langfun.core.coding.python.generation import PythonFunction

from langfun.core.coding.python.correction import correct
from langfun.core.coding.python.correction import run_with_correction
from langfun.core.coding.python.correction import CodeWithError

from langfun.core.coding.python.sandboxing import SandboxOutput
from langfun.core.coding.python.sandboxing import MultiProcessingSandbox

# pylint: enable=g-importing-member
# pylint: enable=g-bad-import-order



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Compositions of different LLM models."""
import random
from typing import Annotated

import langfun.core as lf
import pyglove as pg


@pg.use_init_args(['candidates', 'seed'])
class RandomChoice(lf.LanguageModel):
  """Random choice of a list of LLM models."""

  candidates: Annotated[
      list[lf.LanguageModel],
      (
          'A list of LLMs as candidates to choose from.'
      )
  ]

  seed: Annotated[
      int,
      (
          'The random seed to use for the random choice.'
      )
  ] = 0

  def _on_bound(self):
    super()._on_bound()
    self._rand = random.Random(self.seed)
    # Applying sampling options to all candidates.
    parent_non_default = self.sampling_options.sym_nondefault()
    if parent_non_default:
      for c in self.candidates:
        c.sampling_options.rebind(
            parent_non_default, notify_parents=False, raise_on_no_change=False
        )

  @property
  def model_id(self) -> str:
    model_ids = ', '.join(
        sorted(c.model_id for c in self.candidates)
    )
    return f'RandomChoice({model_ids})'

  @property
  def resource_id(self) -> str:
    resource_ids = ', '.join(
        sorted(c.resource_id for c in self.candidates)
    )
    return f'RandomChoice({resource_ids})'

  def _select_lm(self) -> lf.LanguageModel:
    """Selects a random LLM from the candidates."""
    return self._rand.choice(self.candidates)

  def sample(
      self,
      prompts: list[str | lf.Message],
      *,
      cache_seed: int = 0,
      **kwargs,
  ) -> list[lf.LMSamplingResult]:
    return self._select_lm().sample(
        prompts, cache_seed=cache_seed, **kwargs
    )

  def __call__(
      self, prompt: str | lf.Message, *, cache_seed: int = 0, **kwargs
  ) -> lf.Message:
    return self._select_lm()(prompt, cache_seed=cache_seed, **kwargs)

  def score(
      self,
      prompt: str | lf.Message | list[lf.Message],
      completions: list[str | lf.Message],
      **kwargs,
  ) -> list[lf.LMScoringResult]:
    return self._select_lm().score(prompt, completions, **kwargs)

  def tokenize(
      self,
      prompt: str | lf.Message,
      **kwargs,
  ) -> list[tuple[str | bytes, int]]:
    return self._select_lm().tokenize(prompt, **kwargs)

  def _sample(self, *arg, **kwargs):
    assert False, 'Should never trigger.'



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The base of symbolic mapping methods."""

import functools
import io
from typing import Annotated, Any, Callable
import langfun.core as lf
from langfun.core.structured import schema as schema_lib
import pyglove as pg


class MappingError(Exception):  # pylint: disable=g-bad-exception-name
  """Mapping error."""

  def __init__(self, lm_response: lf.Message, cause: Exception):
    self._lm_response = lm_response
    self._cause = cause

  @property
  def lm_response(self) -> lf.Message:
    """Returns the LM response that failed to be mapped."""
    return self._lm_response

  @property
  def cause(self) -> Exception:
    """Returns the cause of the error."""
    return self._cause

  def __str__(self) -> str:
    return self.format(include_lm_response=True)

  def format(self, include_lm_response: bool = True) -> str:
    """Formats the mapping error."""
    r = io.StringIO()
    error_message = str(self.cause).rstrip()
    r.write(
        pg.colored(
            f'{self.cause.__class__.__name__}: {error_message}', 'magenta'
        )
    )
    if include_lm_response:
      r.write('\n\n')
      r.write(pg.colored('[LM Response]', 'blue', styles=['bold']))
      r.write('\n')
      r.write(pg.colored(self.lm_response.text, 'blue'))
    return r.getvalue()


@pg.use_init_args(['input', 'output', 'schema', 'context'])
class MappingExample(lf.NaturalLanguageFormattable,
                     lf.Component,
                     pg.views.HtmlTreeView.Extension):
  """Mapping example between text, schema and structured value."""

  input: pg.typing.Annotated[
      pg.typing.Any(transform=schema_lib.mark_missing),
      (
          'The input object of the mapping. It could be either a natural '
          'language-based string, or a Python object.'
      ),
  ]

  output: Annotated[
      Any,
      (
          'The output object of the mapping. It could be either a natural '
          'language-based string, or a Python object.'
      ),
  ] = schema_lib.MISSING

  schema: pg.typing.Annotated[
      # Automatic conversion from annotation to schema.
      schema_lib.schema_spec(noneable=True),
      (
          'A `lf.structured.Schema` object that constrains target value '
          'If None, the target is expected to be a natural language-based '
          'response returned from LMs.'
      ),
  ] = lf.contextual(default=None)

  context: Annotated[
      str | None,
      'The natural language context for this mapping. ',
  ] = None

  metadata: Annotated[
      dict[str, Any],
      (
          'The metadata associated with the mapping example, '
          'which chould carry structured data, such as tool function input. '
          'It is a `pg.Dict` object whose keys can be accessed by attributes.'
      ),
  ] = pg.Dict()

  def schema_repr(
      self, protocol: schema_lib.SchemaProtocol = 'python', **kwargs
  ) -> str:
    """Returns the string representation of schema based on protocol."""
    if self.schema is None:
      return ''
    return self.schema.schema_str(protocol, **kwargs)

  @property
  def has_output(self) -> bool:
    """Returns True if the mapping output is present."""
    return self.output != schema_lib.MISSING

  @classmethod
  def value_repr(
      cls,
      value: Any,
      protocol: schema_lib.SchemaProtocol = 'python',
      use_modality_ref: bool = False,
      **kwargs
  ) -> str:
    if isinstance(value, str):
      return value
    if isinstance(value, lf.Modality):
      with lf.modality.format_modality_as_ref():
        return str(value)

    # Placehold modalities if they are present.
    if use_modality_ref and pg.contains(value, type=lf.Modality):
      value = lf.ModalityRef.placehold(value)
    return schema_lib.value_repr(protocol).repr(value, **kwargs)

  def input_repr(
      self,
      protocol: schema_lib.SchemaProtocol = 'python',
      compact: bool = False,
      verbose: bool = True,
      **kwargs
  ) -> str:
    """Returns the string representation of the input object."""
    return self.value_repr(
        self.input, protocol, compact=compact, verbose=verbose, **kwargs
    )

  def output_repr(
      self,
      protocol: schema_lib.SchemaProtocol = 'python',
      compact: bool = False,
      verbose: bool = True,
      **kwargs
  ) -> str:
    """Returns the string representation of the output object."""
    return self.value_repr(
        self.output, protocol, compact=compact, verbose=verbose, **kwargs
    )

  def natural_language_format(self) -> str:
    result = io.StringIO()
    if self.context:
      result.write(pg.colored('[CONTEXT]\n', styles=['bold']))
      result.write(pg.colored(self.context, color='magenta'))
      result.write('\n\n')

    result.write(pg.colored('[INPUT]\n', styles=['bold']))
    result.write(pg.colored(self.input_repr(), color='green'))

    if self.schema is not None:
      result.write('\n\n')
      result.write(pg.colored('[SCHEMA]\n', styles=['bold']))
      result.write(pg.colored(self.schema_repr(), color='red'))

    if schema_lib.MISSING != self.output:
      result.write('\n\n')
      result.write(pg.colored('[OUTPUT]\n', styles=['bold']))
      result.write(pg.colored(self.output_repr(), color='blue'))

    if self.metadata:
      result.write('\n\n')
      result.write(pg.colored('[METADATA]\n', styles=['bold']))
      result.write(pg.colored(str(self.metadata), color='cyan'))
    return result.getvalue().strip()

  @classmethod
  @functools.cache
  def _html_tree_view_config(cls) -> dict[str, Any]:

    def render_value(view, *, value, **kwargs):
      if isinstance(value, lf.Template):
        # Make a shallow copy to make sure modalities are rooted by
        # the input.
        value = value.clone().render()
      if value is None:
        return None
      return view.render(value, **kwargs)

    return pg.views.HtmlTreeView.get_kwargs(
        super()._html_tree_view_config(),
        dict(
            include_keys=['input', 'output', 'context', 'schema', 'metadata'],
            extra_flags=dict(
                render_value_fn=render_value,
            ),
            child_config=dict(
                input=dict(
                    collapse_level=1,
                ),
                output=dict(
                    css_classes=['lf-example-output'],
                    collapse_level=1,
                ),
                schema=dict(
                    css_classes=['lf-example-schema'],
                    collapse_level=1,
                ),
                metadata=dict(
                    css_classes=['lf-example-metadata'],
                    collapse_level=1,
                ),
            ),
        )
    )

  @classmethod
  @functools.cache
  def _html_tree_view_css_styles(cls) -> list[str]:
    return super()._html_tree_view_css_styles() + [
        """
        .lf-example-output {
            color: dodgerblue;
        }
        .lf-example-schema {
            color: blue;
        }
        """
    ]


class Mapping(lf.LangFunc):
  """Base class for mapping.

  {{ preamble }}

  {% if examples -%}
  {% for example in examples -%}
  {{ mapping_template.render(example=example) }}

  {% endfor %}
  {% endif -%}
  {{ mapping_template.render(example=mapping_request) }}
  """

  #
  # Input for mapping.
  #

  input: Annotated[
      pg.Symbolic,
      (
          'The mapping input. It could be `lf.Message` (a pg.Symbolic '
          'subclass) as natural language input, or other symbolic object '
          'as structured input.'
      ),
  ]

  context: Annotated[
      str | None, 'The mapping context. A string as natural language '
  ] = None

  schema: pg.typing.Annotated[
      # Automatic conversion from annotation to schema.
      schema_lib.schema_spec(noneable=True),
      'A `lf.structured.Schema` object that constrains mapping output ',
  ] = None

  permission: Annotated[
      pg.coding.CodePermission,
      'The permission to run the LLM generated code.'
  ] = pg.coding.CodePermission.ASSIGN | pg.coding.CodePermission.CALL

  @property
  def mapping_request(self) -> MappingExample:
    """Returns a MappingExample as the mapping request."""
    if isinstance(self.input, lf.Message):
      input_value = self.input.text
    else:
      input_value = pg.Ref(self.input)
    return MappingExample(
        input=input_value,
        schema=pg.Ref(self.schema),
        context=self.context,
    )

  #
  # Customizable in child classes.
  #

  preamble: Annotated[
      lf.Template,
      'Preamble used as mapping instructions.',
  ]

  mapping_template: Annotated[
      lf.Template | None,
      (
          'Template for demonstrating current mapping based on a '
          'MappingExample object. When the output of the mapping example is '
          'absent, the demonstration represents a mapping request.'
      ),
  ] = lf.Template(
      """
      {%- if example.context -%}
      {{ context_title}}:
      {{ example.context | indent(2, True)}}

      {% endif -%}

      {{ input_title }}:
      {{ example.input_repr(protocol, compact=False) | indent(2, True) }}

      {% if example.schema -%}
      {{ schema_title }}:
      {{ example.schema_repr(protocol) | indent(2, True) }}

      {% endif -%}

      {{ output_title }}:
      {%- if example.has_output %}
      {{ example.output_repr(protocol, compact=False) | indent(2, True) }}
      {% endif -%}
      """
  )

  input_title: Annotated[str, 'The section title for input.'] = 'INPUT'

  output_title: Annotated[str, 'The section title for output.'] = 'OUTPUT'

  context_title: Annotated[str, 'The section title for context.'] = 'CONTEXT'

  schema_title: Annotated[str, 'The section title for schema.'] = 'SCHEMA'

  protocol: Annotated[
      str,
      (
          'A string representing the protocol for formatting the prompt. '
          'Built-in Langfun protocols are: `python` and `json`.'
      ),
  ] = 'python'

  #
  # Other user-provided flags.
  #

  examples: Annotated[
      list[MappingExample] | None,
      'Fewshot examples for improving the quality of mapping.',
  ] = lf.contextual(default=None)

  autofix: Annotated[
      int,
      (
          'Max attempts for LLM-based code correction. '
          'If 0 (default), there is no automatic correction. '
          'This flag is effective only when the output needs to be structured.'
      ),
  ] = 0

  autofix_lm: Annotated[
      lf.LanguageModel,
      (
          'Language model used for code correction. '
          'If None, `lm` will be used. This flag is effective only when the '
          'output needs to be structured.'
      ),
  ] = lf.contextual(default=None)

  default: Annotated[
      Any,
      (
          'The default value to use if the LM response is not a valid code '
          'based on the schema (after autofix). '
          'If unspecified, error will be raisen.'
      ),
  ] = lf.RAISE_IF_HAS_ERROR

  response_postprocess: Annotated[
      Callable[[str], str] | None,
      (
          'A callable object that post process the raw LLM response before '
          'parsing it into the output Python object.'
      )
  ] = None

  #
  # Key methods for implementing specific mappings.
  #

  def transform_input(self, lm_input: lf.Message) -> lf.Message:
    # Find modalities to fill the input message.
    lm_input.metadata.update(
        examples=pg.Ref(self.examples),
        input=pg.Ref(self.input),
        schema=pg.Ref(self.schema) if self.schema is not None else None,
    )
    if isinstance(self.input, lf.Message):
      lm_input.source = self.input
    return lm_input

  def transform_output(self, lm_output: lf.Message) -> lf.Message:
    """Transforms LM response into structure if schema is present."""
    try:
      lm_output = self.postprocess_response(lm_output)
      lm_output.result = self.postprocess_result(self.parse_result(lm_output))
    except Exception as e:  # pylint: disable=broad-exception-caught
      if (self.lm.cache is not None
          and lm_output.lm_input.cache_seed is not None):
        success = self.lm.cache.delete(
            self.lm, lm_output.lm_input, lm_output.lm_input.cache_seed
        )
        assert success
      if self.default == lf.RAISE_IF_HAS_ERROR:
        raise MappingError(lm_output, e) from e
      lm_output.result = self.default
    return lm_output

  def parse_result(self, lm_output: lf.Message) -> Any:
    """Parse result from LLM response."""
    schema = self.mapping_request.schema
    if schema is None:
      return None
    response_text = lm_output.text

    # For Gemini, we might have tool calls in the metadata, use tool call codes
    # to construct the response text if it's present.
    # NOTE(daiyip): This logic is subject to change.
    if 'tool_calls' in lm_output.metadata:
      assert lm_output.metadata['tool_calls'], lm_output.metadata
      response_text = '\n'.join(
          tc.text for tc in lm_output.metadata['tool_calls']
      )
    return schema.parse(
        response_text,
        protocol=self.protocol,
        additional_context=self.globals(),
        autofix=self.autofix,
        autofix_lm=self.autofix_lm or self.lm,
        permission=self.permission,
    )

  def postprocess_response(self, response: lf.Message) -> lf.Message:
    """Post process LLM response."""
    if self.response_postprocess is not None:
      postprocessed_text = self.response_postprocess(response.text)
      if postprocessed_text != response.text:
        return lf.AIMessage(postprocessed_text, source=response)
    return response

  def postprocess_result(self, result: Any) -> Any:
    """Post process structured output."""
    return result

  def globals(self) -> dict[str, Any]:
    """Gets additional symbol definitions besides schema as globals."""
    return {'ModalityRef': lf.modality.ModalityRef}




# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tracking evaluation run progress."""

import os
import langfun.core as lf
from langfun.core.eval.v2 import example as example_lib
from langfun.core.eval.v2 import experiment as experiment_lib
import pyglove as pg

Runner = experiment_lib.Runner
Experiment = experiment_lib.Experiment
Example = example_lib.Example


def progress_tracker(tqdm: bool = False) -> experiment_lib.Plugin:
  """Creates a progress tracker as a plugin.

  Args:
    tqdm: If True, force using tqdm for progress update.

  Returns:
    The progress tracker plugin.
  """
  if tqdm or not lf.console.under_notebook():
    return _TqdmProgressTracker()
  else:
    return _HtmlProgressTracker()


class _HtmlProgressTracker(experiment_lib.Plugin):
  """HTML progress tracker plugin."""

  def on_run_start(
      self,
      runner: Runner,
      root: Experiment
  ) -> None:
    # Display the experiment if running under notebook and not using tqdm.
    assert lf.console.under_notebook()
    with pg.view_options(
        collapse_level=None,
        extra_flags=dict(
            current_run=runner.current_run,
            interactive=True,
        )
    ):
      lf.console.display(runner.current_run.experiment)


ProgressBarId = int


class _TqdmProgressTracker(experiment_lib.Plugin):
  """Tqdm process updater plugin."""

  def _on_bound(self):
    super()._on_bound()
    self._overall_progress: ProgressBarId | None = None
    self._leaf_progresses: dict[str, ProgressBarId] = {}

  def experiment_progress(
      self, experiment: Experiment) -> lf.concurrent.ProgressBar:
    """Returns the progress of the experiment."""
    assert experiment.is_leaf
    return self._leaf_progresses[experiment.id]

  def on_run_start(
      self,
      runner: Runner,
      root: Experiment
  ) -> None:
    """Called when a runner is started."""
    self._overall_progress = lf.concurrent.ProgressBar.install(
        label='All', total=len(root.leaf_nodes), color='blue'
    )
    self._leaf_progresses = {
        leaf.id: lf.concurrent.ProgressBar.install(
            label=f'[#{i + 1} - {leaf.id}]',
            total=(len(runner.current_run.example_ids)
                   if runner.current_run.example_ids else leaf.num_examples),
            color='cyan',
            status=None
        )
        for i, leaf in enumerate(root.leaf_nodes)
    }
    summary_link = Experiment.link(
        os.path.join(runner.current_run.output_root, 'summary.html')
    )
    lf.console.write(f'Summary: {summary_link}.', color='green')

  def on_run_complete(
      self,
      runner: Runner,
      root: Experiment
  ) -> None:
    """Called when a runner is complete."""
    lf.concurrent.ProgressBar.update(
        self._overall_progress,
        color='green',
        status='ALL COMPLETED.',
    )
    lf.concurrent.ProgressBar.uninstall(self._overall_progress)
    self._overall_progress = None
    for progress in self._leaf_progresses.values():
      lf.concurrent.ProgressBar.uninstall(progress)
    self._leaf_progresses = {}

  def on_experiment_start(
      self,
      runner: Runner,
      experiment: Experiment
  ) -> None:
    """Called when an evaluation is started."""

  def on_experiment_skipped(
      self,
      runner: Runner,
      experiment: Experiment
  ) -> None:
    """Called when an evaluation is skipped."""
    if experiment.is_leaf:
      lf.concurrent.ProgressBar.update(
          self.experiment_progress(experiment),
          delta=experiment.progress.num_total,
          status='Skipped.',
      )
      lf.concurrent.ProgressBar.update(
          self._overall_progress,
          status=f'Skipped {experiment.id}.',
      )

  def on_experiment_complete(
      self,
      runner: Runner,
      experiment: Experiment
  ) -> None:
    """Called when an evaluation is complete."""
    if experiment.is_leaf:
      lf.concurrent.ProgressBar.update(
          self.experiment_progress(experiment),
          color='green',
      )
      lf.concurrent.ProgressBar.update(
          self._overall_progress,
          delta=1,
          status=f'{experiment.id} COMPLETED.',
      )

  def on_example_start(
      self,
      runner: Runner,
      experiment: Experiment,
      example: Example
  ) -> None:
    """Called when an evaluation example is started."""

  def on_example_skipped(
      self,
      runner: Runner,
      experiment: Experiment,
      example: Example
  ) -> None:
    """Called when an evaluation example is skipped."""
    del runner, example
    lf.concurrent.ProgressBar.update(
        self.experiment_progress(experiment),
        delta=1,
    )

  def on_example_complete(
      self,
      runner: Runner,
      experiment: Experiment,
      example: Example
  ) -> None:
    """Called when an evaluation example is complete."""
    lf.concurrent.ProgressBar.update(
        self.experiment_progress(experiment),
        delta=1,
        status=self.status(experiment),
    )

  def status(self, experiment: Experiment) -> str:
    """Returns the progress text of the evaluation."""
    items = []
    for metric in experiment.metrics:
      for metric_value in metric.values():
        items.append(
            f'{metric_value.sym_path.key}={metric_value.format(verbose=True)}'
        )
    error_tags = {}
    for entry in experiment.progress.execution_summary.breakdown.values():
      error_tags.update(entry.error_tags)

    if error_tags:
      items.extend(
          [f'{k}={v}' for k, v in error_tags.items() if v > 0]
      )
    return ', '.join(items)



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base class for language models through REST APIs."""

import functools
from typing import Annotated, Any, Callable

import langfun.core as lf
# Placeholder for Google-internal internet access import.
import requests


class REST(lf.LanguageModel):
  """REST-based language model."""

  api_endpoint: Annotated[
      str,
      'The endpoint of the REST API.'
  ]

  request: Annotated[
      Callable[[lf.Message, lf.LMSamplingOptions], dict[str, Any]],
      'A function to convert a Langfun message to a JSON request.'
  ]

  result: Annotated[
      Callable[[dict[str, Any]], lf.LMSamplingResult],
      'A function to convert a JSON response to an LMSamplingResult.'
  ]

  model: Annotated[
      str | None,
      'Model ID.'
  ] = None

  headers: Annotated[
      dict[str, Any] | None,
      'The headers for the REST API.'
  ] = None

  @functools.cached_property
  def _api_initialized(self) -> bool:
    """Returns whether the API is initialized."""
    self._initialize()
    return True

  def _initialize(self) -> None:
    """Initializes the API. Subclasses can override."""

  def session(self) -> requests.Session:
    assert self._api_initialized
    s = self._session()
    # Placeholder for Google-internal session adapter.
    s.headers.update(self.headers or {})
    return s

  def _session(self) -> requests.Session:
    """Creates a new session."""
    return requests.Session()

  def _on_bound(self):
    super()._on_bound()
    self.__dict__.pop('_api_initialized', None)

  def _sample(self, prompts: list[lf.Message]) -> list[lf.LMSamplingResult]:
    assert self._api_initialized
    return self._parallel_execute_with_currency_control(
        self._sample_single, prompts
    )

  def _sample_single(self, prompt: lf.Message) -> lf.LMSamplingResult:
    try:
      with self.session() as session:
        return self._parse_response(
            session.post(
                self.api_endpoint,
                json=self.request(prompt, self.sampling_options),
                timeout=self.timeout,
            )
        )
    except (
        requests.exceptions.Timeout,
        requests.exceptions.ReadTimeout,
        requests.exceptions.ConnectTimeout,
        TimeoutError,
    ) as e:
      raise lf.TemporaryLMError(str(e)) from e
    except (
        requests.exceptions.ConnectionError,
        ConnectionError,
    ) as e:
      error_message = str(e)
      if 'REJECTED_CLIENT_THROTTLED' in error_message:
        raise lf.TemporaryLMError(error_message) from e
      if 'UNREACHABLE_NO_RESPONSE' in error_message:
        raise lf.TemporaryLMError(error_message) from e
      if 'UNREACHABLE_ERROR' in error_message:
        raise lf.TemporaryLMError(error_message) from e
      raise lf.LMError(error_message) from e

  def _error(self, status_code: int, content: str) -> lf.LMError:
    if status_code == 429:
      error_cls = lf.RateLimitError
    elif status_code in (
        500,  # Server side issue (might be bug).
        502,  # Bad gateway (upstream issue, might retry).
        503,  # Servers currently under load, retry after a brief wait.
        529,  # Overloaded, retry after a brief wait.
        499,  # Client Closed Request
    ):
      error_cls = lf.TemporaryLMError
    else:
      error_cls = lf.LMError
    return error_cls(f'{status_code}: {content}')

  def _parse_response(self, response: requests.Response) -> lf.LMSamplingResult:
    """Parses Anthropic's response."""
    if response.status_code == 200:
      return self.result(response.json())
    else:
      raise self._error(response.status_code, response.content)

  @property
  def max_concurrency(self) -> int | None:
    """Returns the max concurrency for this model."""
    rate_limits = self.model_info.rate_limits
    if rate_limits is not None:
      return self.estimate_max_concurrency(
          max_requests_per_minute=rate_limits.max_requests_per_minute,
          max_tokens_per_minute=rate_limits.max_tokens_per_minute
      )
    return None



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import unittest
import langfun.core as lf
from langfun.core.llms import fake
from langfun.core.structured import tokenization
import pyglove as pg


class Answer(pg.Object):
  result: int


class TokenizationTest(unittest.TestCase):

  def test_bad_call(self):

    with self.assertRaisesRegex(ValueError, '`lm` must be specified'):
      tokenization.tokenize('hi')

  def test_tokenize(self):
    self.assertEqual(
        tokenization.tokenize('hi', lm=fake.Echo()),
        [('hi', 0)]
    )

  def test_atokenize(self):
    with lf.context(lm=fake.Echo()):
      self.assertEqual(
          asyncio.run(tokenization.atokenize('hi')),
          [('hi', 0)]
      )

  def test_tokenize_with_lm_from_the_context(self):
    with lf.context(lm=fake.Echo()):
      self.assertEqual(
          tokenization.tokenize('hi'),
          [('hi', 0)]
      )


if __name__ == '__main__':
  unittest.main()



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base class for Langfun evaluation tasks."""

import abc
import datetime
import functools
import threading
import time

from typing import Annotated, Any, Callable, Iterable
import langfun.core as lf
import langfun.core.coding as lf_coding

from langfun.core.eval.v2 import example as example_lib
from langfun.core.eval.v2 import experiment as experiment_lib
from langfun.core.eval.v2 import metric_values as metric_values_lib
from langfun.core.eval.v2 import metrics as metrics_lib

import pyglove as pg


class Evaluation(experiment_lib.Experiment):
  """Evaluation.

  An evaluation can be a leaf node or a container of other evaluations,
  depending on whether the current evaluation object is configured with
  any `pg.oneof`.

  For example, `MyEval(lm=pg.oneof([lf.llms.Gpt4(), lf.llms.Gemini1_5Pro()]))`
  is a container of two sub-experiments, one for each LLM. In such case, the
  evaluation object with `pg.oneof` is called a hyper evaluation, which
  represents a search space of evaluations, and each sub-evaluation is called
  a leaf evaluation, which will perform the actual evaluation.
  """

  inputs: Annotated[
      pg.Functor,
      'A functor that returns a list of inputs.'
  ]

  metrics: Annotated[
      list[metrics_lib.Metric],
      'The metrics to be evaluated.'
  ]

  max_workers: Annotated[
      int,
      'The maximum number of workers to use for the evaluation.'
  ] = 32

  def _on_bound(self):
    # Invalidate cached properties.
    self.__dict__.pop('is_leaf', None)
    self.__dict__.pop('children', None)
    super()._on_bound()
    self._log_entries = []
    self._log_lock = threading.Lock()

  #
  # Handling evaluation hierarchy (materialized vs. hyper evaluations).
  #

  @functools.cached_property
  def is_leaf(self) -> bool:
    """Returns whether the task is a leaf."""
    return self.is_deterministic

  @functools.cached_property
  def children(self) -> list['Evaluation']:
    """Returns the children tasks."""
    if self.is_leaf:
      return []
    children = []
    for i, child in enumerate(pg.iter(self)):
      child.sym_setparent(self)
      child.sym_setpath(self.sym_path + 'children' + i)
      children.append(child)
    return children

  #
  # Handling evaluation inputs.
  #

  @functools.cached_property
  def example_inputs(self) -> Iterable[Any]:
    """Returns the examples from the inputs."""
    return self.inputs()

  def example_input_by_id(self, example_id: int) -> Any:
    """Returns the example from the inputs by ID."""
    assert example_id <= len(self.example_inputs), example_id
    return self._example_input_by_id[example_id]

  @functools.cached_property
  def _example_input_by_id(self) -> dict[int, Any]:
    """Returns the examples from the inputs by ID."""
    return {i + 1: v for i, v in enumerate(self.example_inputs)}

  @property
  def num_examples(self) -> int:
    """Returns the number of examples from the inputs."""
    # NOTE(daiyip): setting `num_examples` of the input functor allows fast
    # retrieval of number of examples without iterating the whole dataset.
    num_examples = getattr(self.inputs, 'num_examples', None)
    if not isinstance(num_examples, int):
      it = self.example_inputs
      if hasattr(it, '__len__'):
        num_examples = len(it)
      else:
        num_examples = len(list(it))
    return num_examples

  #
  # Evaluation logics.
  #

  @abc.abstractmethod
  def process(
      self,
      example: example_lib.Example
  ) -> Any | tuple[Any, dict[str, Any]]:
    """Processes a single example from the evaluation set.

    Users should override this method to implement the evaluation logic.

    Args:
      example: An example object to process. `example.input` is an object
        returned from `Evaluable.inputs`.

    Returns:
      A processed output. Or a tuple of (output, metadata).
      The output will be used for computing the metrics, and the metadata will
      be included in the evaluation HTML view.
    """

  def evaluate(
      self,
      example: example_lib.Example | int,
      raise_if_has_error: bool = False,
      reevaluate_upon_previous_errors: bool = True,
  ) -> example_lib.Example:
    """Evaluates a single example input.

    Args:
      example: An example ID or an example object with ID.
      raise_if_has_error: Whether to raise an error if the example has error.
      reevaluate_upon_previous_errors: Whether to reevaluate the example if
        the previous checkpointed run has error.

    Returns:
      The evaluated example with the output and metric metadata populated.
    """
    if isinstance(example, int):
      example = example_lib.Example(id=example)
    assert isinstance(example, example_lib.Example), example

    if pg.MISSING_VALUE == example.input:
      example.input = self.example_input_by_id(example.id)

    checkpointed = self._state.ckpt_example(example.id)

    with pg.timeit('evaluate') as timeit, lf.track_usages() as usage_summary:
      if checkpointed is None or (
          reevaluate_upon_previous_errors and checkpointed.has_error
      ):
        if checkpointed is None:
          self.info(
              f'Example {example.id} is being processed for the first time '
              'as no prior run is found. '
          )
        else:
          self.info(
              f'Example {example.id} is being reprocessed as prior run '
              f'contains error: {checkpointed.error}'
          )
        example.start_time = time.time()
        self._state.update(example, in_progress=True)
        self._process(example, raise_if_has_error=raise_if_has_error)
      else:
        if checkpointed.has_error:
          self.warning(
              f'Example {example.id} skipped processing despite its checkpoint '
              f'has error: {checkpointed.error}, as '
              '`reevaluate_upon_previous_errors` is set to False.'
          )
        else:
          self.info(
              f'Example {example.id} skipped processing as prior run '
              'is available and error free.'
          )
        example.start_time = checkpointed.start_time
        self._state.update(example, in_progress=True)

        # Use the output and metadata obtained from the previous processing.
        example.output = checkpointed.output
        example.metadata = checkpointed.metadata
        example.error = checkpointed.error
        example.newly_processed = False
        example.execution_status = checkpointed.execution_status
        example.end_time = checkpointed.end_time

        # For previously processed examples, we merge previous usages as
        # cached, so the usage summary will account previous usages, but as
        # cached.
        assert checkpointed.usage_summary is not None
        usage_summary.merge(checkpointed.usage_summary, as_cached=True)

      # Recompute the metrics and metadata for the example even its processed
      # output and metadata were from the cache.
      # NOTE(daiyip): It's possible that metrics could use LLMs, so we need to
      # track the usage of the metrics separately.
      with pg.timeit('metric'):
        self.info(f'Starting metric computation for example {example.id}.')
        metric_metadata = {}
        for metric in self.metrics:
          metric_metadata.update(metric.audit(example))
        example.metric_metadata = metric_metadata
        self.info(f'Completed metric computation for example {example.id}.')

    # For previously processed examples, we keep the execution status for the
    # processing step.
    execution_status = dict(example.execution_status or {})
    execution_status.update(timeit.status())

    example.execution_status = execution_status
    example.usage_summary = usage_summary
    if example.newly_processed:
      example.end_time = time.time()

    self._state.update(example, in_progress=False)
    return example

  def _process(
      self,
      example: example_lib.Example,
      raise_if_has_error: bool = False
  ) -> None:
    """Processes a single example."""
    with (
        pg.notify_on_change(False),
        pg.allow_writable_accessors(True),
        # NOTE(daiyip): set the `input` symbol of the globals to None, so
        # LLM generated code with calls to `input` will raise an error, thus
        # not blocking the evaluation.
        lf_coding.context(input=None),
    ):
      try:
        with pg.timeit('process'):
          output = self.process(example)
        if (isinstance(output, tuple)
            and len(output) == 2
            and isinstance(output[1], dict)):
          output, metadata = output
        else:
          metadata = {}
        example.output = output
        example.metadata = metadata
      except BaseException as e:  # pylint: disable=broad-except
        if raise_if_has_error:
          raise
        example.error = pg.ErrorInfo.from_exception(e)

  #
  # Handling evaluation scheduling.
  #

  def resource_ids(self) -> set[str]:
    """Returns a set of resource IDs required by this evaluation.

    Resource IDs are used to by the runner to determine which evaluations can
    be run in parallel. Evaluations using the same resource key will be run
    sequentially.

    Returns:
      A unique string representing the resource required.
    """
    return {
        v.resource_id for _, v in self.sym_init_args.items()
        if isinstance(v, lf.LanguageModel)
    }

  #
  # Handling evaluation state.
  #

  @property
  def state(self) -> 'EvaluationState':
    """Returns the state of the evaluation."""
    return self._state

  def load_state(
      self,
      state_file: str,
      *,
      load_example_metadata: bool = True,
      filter: Callable[[example_lib.Example], bool] | None = None,  # pylint: disable=redefined-builtin
      raise_if_not_exist: bool = False
  ) -> None:
    """Loads saved state from a sequence IO file."""
    if pg.io.path_exists(state_file):
      self._state.load(
          state_file,
          example_input_by_id=self.example_input_by_id,
          load_example_metadata=load_example_metadata,
          filter=filter,
      )
    elif raise_if_not_exist:
      raise ValueError(f'State file {state_file} does not exist.')

  def _reset(self) -> None:
    """Resets the state of the evaluation."""
    super()._reset()
    if self.is_leaf:
      # Create a new state for the leaf evaluation.
      self._state = EvaluationState()
      for metric in self.metrics:
        metric.reset()

  #
  # Evaluation-level logging.
  #

  def _log(self, log_func, level: lf.logging.LogLevel, message: str, **kwargs):
    # Write to external logging system.
    log_message = message
    if self.id not in log_message:
      log_message = f'{self.id}: {log_message}'
    if kwargs:
      log_message = f'{log_message} (metadata: {kwargs!r})'
    log_func(log_message)

    # Add to experiment log history.
    log_entry = lf.logging.LogEntry(
        level=level,
        time=datetime.datetime.now(),
        message=message,
        metadata=kwargs,
    )
    with self._log_lock:
      self._log_entries.append(log_entry)

  def debug(self, message: str, **kwargs):
    """Logs a debug message to the session."""
    self._log(pg.logging.debug, 'debug', message, **kwargs)

  def info(self, message: str, **kwargs):
    """Logs an info message to the session."""
    self._log(pg.logging.info, 'info', message, **kwargs)

  def warning(self, message: str, **kwargs):
    """Logs a warning message to the session."""
    self._log(pg.logging.warning, 'warning', message, **kwargs)

  def error(self, message: str, **kwargs):
    """Logs an error message to the session."""
    self._log(pg.logging.error, 'error', message, **kwargs)

  def fatal(self, message: str, **kwargs):
    """Logs a fatal message to the session."""
    # We use error level for fatal message, which does not trigger assertion.
    self._log(pg.logging.error, 'fatal', message, **kwargs)

  #
  # HTML views.
  #

  def _html_tree_view_content(
      self, *, view, extra_flags: dict[str, Any] | None, **kwargs
  ):
    if not self.is_leaf:
      return super()._html_tree_view_content(
          view=view, extra_flags=extra_flags, **kwargs
      )

    extra_flags = extra_flags or {}
    run = extra_flags.pop('current_run', None)
    if extra_flags.pop('card_view', True):
      return self._summary_card_view(
          extra_flags.get('interactive', True), run
      )
    assert run is not None
    return self._details_view(run)

  def _parameter_badge(self, key, value) -> pg.Html.WritableTypes:
    """Renders a badge for a parameter."""
    face_value = pg.format(
        value,
        compact=True,
        python_format=True,
        hide_default_values=True,
        use_inferred=True
    )
    short_text = face_value
    if len(face_value) > 40:
      short_text = f'{type(value).__name__}(...)'
    label = f'{key.split(".")[-1]}: {short_text}'
    tooltip = f'{key}: {face_value}'
    return pg.views.html.controls.Badge(
        text=label,
        tooltip=tooltip,
        css_classes=['parameter'],
        interactive=False,
    )

  def _summary_card_view(
      self,
      interactive: bool = True,
      run: experiment_lib.Run | None = None,
  ) -> pg.Html.WritableTypes:
    """Renders the summary card view of the evaluation."""
    del run
    return pg.Html(
        pg.Html.element(
            'div',
            [
                pg.views.html.controls.LabelGroup([
                    self._parameter_badge(k, v)
                    for k, v in self.non_default_values(
                        flatten=True
                    ).items()
                ], css_classes=['parameter-group']),
                pg.Html.element(
                    'div',
                    [
                        m.to_html(
                            extra_flags=dict(
                                interactive=interactive,
                            )
                        )
                        for m in self.metrics
                    ],
                    css_classes=['metric-group'],
                ),
            ],
            css_classes=['badge-groups'],
        )
    )

  def _details_view(
      self, run: experiment_lib.Run
  ) -> pg.Html:
    """Renders the details view of the evaluation."""

    def _title():
      additional_links = []
      return pg.Html.element(
          'div',
          [
              pg.views.html.controls.LabelGroup(
                  [
                      pg.views.html.controls.Label(
                          'Summary',
                          link=run.output_link('summary.html'),
                          css_classes=['summary-link'],
                      ),
                      '|',
                      pg.views.html.controls.Label(
                          'Directory',
                          link=self.output_link(run, ''),
                          css_classes=['dir-link'],
                      ),
                  ] + additional_links,
                  css_classes=['experiment-links'],
              ),
              pg.views.html.controls.Label(
                  self.id,
                  css_classes=['experiment-id'],
              ),
              self.progress.to_html(
                  extra_flags=dict(interactive=False),
              ),
              self.usage_summary.to_html(
                  extra_flags=dict(as_badge=True, interactive=False),
              ),
          ]
      )

    def _parameter_badges():
      """Renders a tab group for a metric (group)."""
      return pg.views.html.controls.LabelGroup(
          [
              self._parameter_badge(k, v)
              for k, v in self.non_default_values(flatten=True).items()
          ],
          css_classes=['parameter-group'],
      )

    def _definition_tab() -> pg.views.html.controls.Tab:
      """Renders a tab for the definition of the evaluation."""
      return pg.views.html.controls.Tab(
          label='Definition',
          content=pg.Html.element(
              'div',
              [
                  pg.views.html.controls.Label(
                      pg.format(
                          self,
                          compact=False,
                          verbose=False,
                          use_inferred=True,
                          hide_frozen=True,
                          exclude_keys=set(['progress', 'usage_summary'])
                      ),
                      css_classes=['eval-definition'],
                  ),
              ]
          )
      )

    def _in_progress_tab() -> pg.views.html.controls.Tab | None:
      """Renders a tab for the in progress examples."""
      if not self.state.in_progress_examples:
        return None
      return pg.views.html.controls.Tab(
          label='In Progress',
          content=pg.Html.element(
              'div', [
                  self._in_progress_view(
                      list(self.state.in_progress_examples.values())
                  )
              ]
          )
      )

    def _metric_tab(metric: metrics_lib.Metric) -> pg.views.html.controls.Tab:
      """Renders a tab for a metric (group)."""
      return pg.views.html.controls.Tab(
          label=f'Metric: {metric.name}',
          content=pg.Html.element(
              'div',
              [
                  metric.to_html(
                      extra_flags=dict(
                          interactive=False,
                      )
                  ),
                  pg.views.html.controls.TabControl(
                      tabs=[
                          _metric_value_tab(mv)
                          for mv in metric.values()
                      ]
                  )
              ]
          )
      )

    def _metric_value_tab(
        metric_value: metric_values_lib.MetricValue
    ) -> pg.views.html.controls.Tab:
      """Renders the example links for a metric value."""
      return pg.views.html.controls.Tab(
          label=metric_value.sym_path.key,
          content=pg.Html.element(
              'div',
              [
                  pg.views.html.controls.Label(
                      str(dp.example_id),
                      link=self.output_link(run, f'{dp.example_id}.html'),
                      target='example-view',
                      css_classes=['example-link'],
                  )
                  for dp in sorted(
                      metric_value.data_points, key=lambda dp: dp.example_id
                  )
              ]
          )
      )

    def _logs_tab() -> pg.views.html.controls.Tab:
      """Renders a tab for the logs of the evaluation."""
      with self._log_lock:
        log_history = '\n'.join(str(l) for l in self._log_entries)
      return pg.views.html.controls.Tab(
          label='Logs',
          content=pg.Html.element(
              'div',
              [
                  pg.Html.element(
                      'textarea',
                      [pg.Html.escape(log_history)],
                      readonly=True,
                      css_classes=['logs-textarea'],
                  )
              ]
          )
      )

    def _main_tabs() -> pg.Html:
      return pg.Html.element(
          'div',
          [
              pg.views.html.controls.TabControl(
                  [
                      _definition_tab(),
                      [_metric_tab(m) for m in self.metrics],
                      _in_progress_tab(),
                      _logs_tab(),
                  ],
                  selected=1,
              )
          ],
      )

    return pg.Html.element(
        'div',
        [
            _title(),
            _parameter_badges(),
            _main_tabs(),
            pg.Html.element(
                'iframe', [],
                name='example-view',
                src='about:blank',
                title='Example view.',
                css_classes=['example-view'],
            ),
        ],
        css_classes=['eval-details'],
    )

  def _in_progress_view(
      self, in_progress_examples: list[example_lib.Example]
  ) -> pg.Html:
    """Renders a HTML view for the in-progress examples."""
    current_time = time.time()
    logs = [f'(Total {len(in_progress_examples)} examples in progress)']
    for example in in_progress_examples:
      if example.newly_processed:
        logs.append(
            f'Example {example.id}: In progress for '
            f'{current_time - example.start_time:.2f} seconds.'
        )
      else:
        logs.append(f'Example {example.id}: Recomputing metrics...')
    return pg.Html.element(
        'textarea',
        [pg.Html.escape('\n'.join(logs))],
        readonly=True,
        css_classes=['logs-textarea'],
    )

  def _html_tree_view_config(self) -> dict[str, Any]:
    return dict(
        css_classes=['eval-card'] if self.is_leaf else None
    )

  def _html_tree_view_css_styles(self) -> list[str]:
    return super()._html_tree_view_css_styles() + [
        """
        details.eval-card {
          display: inline-block;
          border: 0px;
          box-shadow: rgba(0, 0, 0, 0.16) 0px 1px 4px;
          margin: 15px;
        }
        .eval-card details {
          border: 0px;
        }
        .badge-groups {
          font-weight: normal;
          padding: 5px;
        }
        .parameter-group {
          display: inline-grid;
          grid-template-rows: auto auto;
          border: 0px;
          margin-right: 10px;
        }
        .parameter.badge {
          margin: 2px;
        }
        .metric-group {
          display: inline-grid;
          grid-template-rows: auto auto;
        }
        .eval-details .progress-bar > .shade {
          visibility: hidden;
          width: 0px;
          margin: 0px;
        }
        .eval-details .progress-label {
          font-size: 16px;
          background-color: #eee;
        }
        .eval-details .progress-time {
          font-size: 16px;
          color: dodgerblue;
          background-color: #eee;
          margin-right: 10px;
        }
        .eval-details .usage-summary.badge {
          color: orange;
          font-size: 16px;
          background-color: #eee;
        }
        .eval-details .experiment-links {
          display: block;
          border: 0px;
          margin: 0px;
        }
        .eval-details .tab-control {
          width: 100%;
        }
        .eval-details .tab-button {
          font-size: large;
        }
        .experiment-links .label {
          color: revert;
          margin: 0px;
          padding: 2px;
        }
        .eval-details .experiment-id {
          font-size: 2.0em;
          font-weight: bold;
          display: block;
        }
        .eval-details .parameter-group {
          display: inline-block;
          padding: 5px;
        }
        .eval-definition {
          white-space: pre;
          background-color: #eee;
          padding: 15px;
        }
        .eval-details .metric-container {
          display: block;
          padding: 15px 0px;
        }
        .example-link {
          color: revert;
        }
        .example-view {
          border: 0px;
          width:100%;
          height:100%;
        }
        .logs-textarea {
          width: 100%;
          height: 500px;
          padding: 5px;
          border: 1px solid #DDD;
          background-color: #EEE;
          resize: vertical;
        }
        """
    ]


class EvaluationState:
  """Evaluation state."""

  class ExampleStatus(pg.Object):
    """Example state."""
    evaluated: Annotated[
        bool,
        'Whether the example is evaluated.'
    ] = False

    in_progress: Annotated[
        bool,
        (
            'Whether the example is in progress. '
        )
    ] = False

    newly_processed: Annotated[
        bool,
        (
            'Whether the example is newly processed. '
            'Applicable only when evaluated is True.'
        )
    ] = False

    has_error: Annotated[
        bool,
        (
            'Whether the example has error. '
            'Applicable only when evaluated is True.'
        )
    ] = False

  def __init__(self):
    super().__init__()
    self._ckpt_examples: dict[int, example_lib.Example] = {}
    self._evaluation_status: dict[
        int, EvaluationState.ExampleStatus
    ] = {}
    self._in_progress_examples: dict[int, example_lib.Example] = {}

  def load(
      self,
      state_file: str,
      *,
      example_input_by_id: Callable[[int], Any] | None = None,
      load_example_metadata: bool | Callable[
          [example_lib.Example], bool] = True,
      filter: Callable[[example_lib.Example], bool] | None = None,  # pylint: disable=redefined-builtin
  ) -> None:
    """Loads the state from the example sequence file."""
    for example in example_lib.Example.iter_ckpts(
        state_file,
        example_input_by_id=example_input_by_id,
        load_example_metadata=load_example_metadata,
    ):
      if filter is not None and not filter(example):
        continue
      example.newly_processed = False
      self._ckpt_examples[example.id] = example

  @property
  def evaluation_status(self) -> dict[int, ExampleStatus]:
    """Returns the evaluation status of the examples."""
    return self._evaluation_status

  @property
  def in_progress_examples(self) -> dict[int, example_lib.Example]:
    """Returns the in-progress examples."""
    return self._in_progress_examples

  @property
  def ckpt_examples(self) -> dict[int, example_lib.Example]:
    """Returns the unevaluated examples from checkpoints."""
    return self._ckpt_examples

  def ckpt_example(self, example_id: int) -> example_lib.Example | None:
    """Returns the unevaluated example from checkpoints for a given ID."""
    return self._ckpt_examples.get(example_id)

  def get_status(self, example_id: int) -> ExampleStatus:
    """Returns the evaluation status of the example."""
    return self._evaluation_status.get(
        example_id, EvaluationState.ExampleStatus()
    )

  def update(self, example: example_lib.Example, in_progress: bool) -> None:
    """Updates the state with the given example."""
    self._update_status(example, in_progress)

    if in_progress:
      self._in_progress_examples[example.id] = example
    else:
      self._in_progress_examples.pop(example.id, None)
      # Processed examples will be removed once it's done.
      self._ckpt_examples.pop(example.id, None)

  def _update_status(
      self,
      example: example_lib.Example,
      in_progress: bool
  ) -> None:
    """Updates the evaluation status of the example."""
    self._evaluation_status[example.id] = (
        EvaluationState.ExampleStatus(
            evaluated=example.output != pg.MISSING_VALUE,
            in_progress=in_progress,
            newly_processed=example.newly_processed,
            has_error=example.has_error,
        )
    )

# Register the logging functions to skip the source of the logging functions.
pg.logging.register_frame_to_skip([
    Evaluation._log,    # pylint: disable=protected-access
    Evaluation.debug,
    Evaluation.info,
    Evaluation.warning,
    Evaluation.error,
    Evaluation.fatal
])



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Language models from Google GenAI."""

import functools
import os
from typing import Annotated, Literal

import langfun.core as lf
from langfun.core.llms import gemini
import pyglove as pg


@lf.use_init_args(['model'])
@pg.members([('api_endpoint', pg.typing.Str().freeze(''))])
class GenAI(gemini.Gemini):
  """Language models provided by Google GenAI."""

  model: pg.typing.Annotated[
      pg.typing.Enum(
          pg.MISSING_VALUE,
          [
              m.model_id for m in gemini.SUPPORTED_MODELS
              if m.provider == 'Google GenAI' or (
                  isinstance(m.provider, pg.hyper.OneOf)
                  and 'Google GenAI' in m.provider.candidates
              )
          ]
      ),
      'The name of the model to use.',
  ]

  api_key: Annotated[
      str | None,
      (
          'API key. If None, the key will be read from environment variable '
          "'GOOGLE_API_KEY'. "
          'Get an API key at https://ai.google.dev/tutorials/setup'
      ),
  ] = None

  api_version: Annotated[
      Literal['v1beta', 'v1alpha'],
      'The API version to use.'
  ] = 'v1beta'

  @functools.cached_property
  def model_info(self) -> lf.ModelInfo:
    return super().model_info.clone(
        override=dict(provider='Google GenAI')
    )

  def session(self):
    assert self._api_initialized
    s = self._session()
    s.headers.update(self.headers or {})
    return s

  @property
  def api_endpoint(self) -> str:
    api_key = self.api_key or os.environ.get('GOOGLE_API_KEY', None)
    if not api_key:
      raise ValueError(
          'Please specify `api_key` during `__init__` or set environment '
          'variable `GOOGLE_API_KEY` with your Google Cloud API key. '
          'Check out '
          'https://cloud.google.com/api-keys/docs/create-manage-api-keys '
          'for more details.'
      )
    return (
        f'https://generativelanguage.googleapis.com/{self.api_version}'
        f'/models/{self.model}:generateContent?'
        f'key={api_key}'
    )


# pylint: disable=invalid-name

#
# Experimental models.
#


class Gemini25FlashImagePreview(GenAI):
  """Gemini 2.5 Flash Image Preview model."""
  model = 'gemini-2.5-flash-image-preview'


class Gemini25Pro(GenAI):
  """Gemini 2.5 Pro GA model."""

  model = 'gemini-2.5-pro'


class Gemini25Flash(GenAI):
  """Gemini 2.5 Flash GA model."""

  model = 'gemini-2.5-flash'


class Gemini2ProExp_20250205(GenAI):
  """Gemini 2.0 Pro experimental model launched on 02/05/2025."""
  model = 'gemini-2.0-pro-exp-02-05'


class Gemini2FlashThinkingExp_20250121(GenAI):
  """Gemini 2.0 Flash Thinking model launched on 01/21/2025."""
  api_version = 'v1beta'
  model = 'gemini-2.0-flash-thinking-exp-01-21'
  timeout = None


class GeminiExp_20241206(GenAI):
  """Gemini Experimental model launched on 12/06/2024."""
  model = 'gemini-exp-1206'


#
# Production models.
#
class Gemini25ProPreview_20250605(GenAI):
  """Gemini 2.5 Pro model launched on 06/05/2025."""
  model = 'gemini-2.5-pro-preview-06-05'


class Gemini25FlashPreview_20250520(GenAI):
  """Gemini 2.5 Flash model launched on 05/20/2025."""
  model = 'gemini-2.5-flash-preview-05-20'


class Gemini25ProPreview_20250506(GenAI):
  """Gemini 2.5 Pro model launched on 05/06/2025."""

  model = 'gemini-2.5-pro-preview-05-06'


class Gemini25FlashPreview_20250417(GenAI):
  """Gemini 2.5 Flash model launched on 04/17/2025."""

  model = 'gemini-2.5-flash-preview-04-17'


class Gemini25ProPreview_20250325(GenAI):
  """Gemini 2.5 Pro model launched on 03/25/2025."""
  model = 'gemini-2.5-pro-preview-03-25'


class Gemini2Flash(GenAI):
  """Gemini 2.0 Flash model (latest stable)."""
  model = 'gemini-2.0-flash'


class Gemini2Flash_001(GenAI):
  """Gemini 2.0 Flash model launched on 02/05/2025."""
  model = 'gemini-2.0-flash-001'


class Gemini2FlashLitePreview_20250205(GenAI):
  """Gemini 2.0 Flash lite preview model launched on 02/05/2025."""
  model = 'gemini-2.0-flash-lite-preview-02-05'


class Gemini15Pro(GenAI):
  """Gemini 1.5 Pro latest stable model."""
  model = 'gemini-1.5-pro'


class Gemini15Pro_002(GenAI):
  """Gemini 1.5 Pro stable version 002."""
  model = 'gemini-1.5-pro-002'


class Gemini15Pro_001(GenAI):
  """Gemini 1.5 Pro stable version 001."""
  model = 'gemini-1.5-pro-001'


class Gemini15Flash(GenAI):
  """Gemini 1.5 Flash latest model."""
  model = 'gemini-1.5-flash'


class Gemini15Flash_002(GenAI):
  """Gemini 1.5 Flash model stable version 002."""
  model = 'gemini-1.5-flash-002'


class Gemini15Flash_001(GenAI):
  """Gemini 1.5 Flash model stable version 001."""
  model = 'gemini-1.5-flash-001'


class Gemini15Flash8B(GenAI):
  """Gemini 1.5 Flash 8B modle (latest stable)."""
  model = 'gemini-1.5-flash-8b'


class Gemini15Flash8B_001(GenAI):
  """Gemini 1.5 Flash 8B model (version 001)."""
  model = 'gemini-1.5-flash-8b-001'


# For backward compatibility.
GeminiPro1_5 = Gemini15Pro
GeminiFlash1_5 = Gemini15Flash

# pylint: enable=invalid-name


def _genai_model(model: str, *args, **kwargs) -> GenAI:
  model = model.removeprefix('google_genai://')
  return GenAI(model=model, *args, **kwargs)


def _register_genai_models():
  """Register GenAI models."""
  for m in gemini.SUPPORTED_MODELS:
    lf.LanguageModel.register('google_genai://' + m.model_id, _genai_model)

_register_genai_models()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for langfun.core.templates.selfplay."""


import unittest
import langfun.core as lf
from langfun.core.templates import selfplay


class NumberGuesser(lf.language_model.LanguageModel):
  guesses: list[int]

  def _on_bound(self):
    super()._on_bound()
    self._index = 0

  def _sample(self, prompts):
    results = []
    for _ in range(len(prompts)):
      guess = self.guesses[self._index]
      self._index = (self._index + 1) % len(self.guesses)
      results.append(lf.LMSamplingResult([lf.LMSample(str(guess))]))
    return results


class NumberGuess(selfplay.SelfPlay):
  target_num: int

  def next_turn(self, last_response):
    if last_response is None:
      return 'Let us get started. Tell me your first guess.'
    guess = int(last_response.text)
    if guess > self.target_num:
      return 'Too large'
    elif guess < self.target_num:
      return 'Too small'
    else:
      return None


class SelfPlayTest(unittest.TestCase):

  def test_play(self):
    g = NumberGuess(target_num=10)

    with lf.context(lm=NumberGuesser(guesses=[50, 20, 5, 10])):
      self.assertEqual(
          g(),
          lf.AIMessage(
              '10', score=0.0, logprobs=None, is_cached=False,
              usage=lf.UsageNotAvailable()
          )
      )

    self.assertEqual(g.num_turns, 4)

  def test_play_with_num_turns(self):
    g = NumberGuess(target_num=10, max_turns=10)

    with lf.context(lm=NumberGuesser(guesses=[50, 20, 5, 2, 5, 4])):
      self.assertEqual(
          g(),
          lf.AIMessage(
              '2', score=0.0, logprobs=None, is_cached=False,
              usage=lf.UsageNotAvailable()
          )
      )

    self.assertEqual(g.num_turns, 10)


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest
from langfun.core.llms import llama_cpp


class LlamaCppRemoteTest(unittest.TestCase):
  """Tests for the LlamaCppRemote model."""

  def test_basics(self):
    lm = llama_cpp.LlamaCppRemote("http://127.0.0.1:8080")
    self.assertEqual(lm.api_endpoint, "http://127.0.0.1:8080/completion")
    self.assertEqual(lm.model_id, "LLaMAC++()")
    lm = llama_cpp.LlamaCppRemote("xxx", model="x")
    self.assertEqual(lm.model_id, "LLaMAC++(x)")


if __name__ == "__main__":
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for structured description."""

import inspect
import unittest

from langfun.core.llms import fake
from langfun.core.structured import description as description_lib
from langfun.core.structured import mapping
import pyglove as pg


class Activity(pg.Object):
  description: str


class Itinerary(pg.Object):
  day: pg.typing.Int[1, None]
  type: pg.typing.Enum['daytime', 'nighttime']
  activities: list[Activity]
  hotel: pg.typing.Str['.*Hotel'] | None


class DescribeStructureTest(unittest.TestCase):

  def test_render(self):
    l = description_lib._DescribeStructure(
        input=Itinerary(
            day=1,
            type='daytime',
            activities=[
                Activity('Visit Golden Gate Bridge.'),
                Activity("Visit Fisherman's Wharf."),
                Activity('Visit Alcatraz Island.'),
            ],
            hotel=None,
        ),
        context='1 day itinerary to SF',
        examples=[
            mapping.MappingExample(
                context='Compute 1 + 2',
                input=3,
                output='The result of 1 + 2 is 3',
            ),
            mapping.MappingExample(
                context='Best activity to do in New York city.',
                input=Activity('Visit Broadway threatres for shows'),
                output=(
                    'The best thing to do in New York city is to watch shows '
                    'in Broadway threatres.'
                ),
            ),
        ],
    )

    self.assertEqual(
        l.render().text,
        inspect.cleandoc("""
            Please help describe PYTHON_OBJECT in natural language.

            INSTRUCTIONS:
              1. Do not add details which are not present in the object.
              2. If a field in the object has None as its value, do not mention it.

            CONTEXT_FOR_DESCRIPTION:
              Compute 1 + 2

            PYTHON_OBJECT:
              ```python
              3
              ```

            NATURAL_LANGUAGE_TEXT:
              The result of 1 + 2 is 3

            CONTEXT_FOR_DESCRIPTION:
              Best activity to do in New York city.

            PYTHON_OBJECT:
              ```python
              Activity(
                description='Visit Broadway threatres for shows'
              )
              ```

            NATURAL_LANGUAGE_TEXT:
              The best thing to do in New York city is to watch shows in Broadway threatres.


            CONTEXT_FOR_DESCRIPTION:
              1 day itinerary to SF

            PYTHON_OBJECT:
              ```python
              Itinerary(
                day=1,
                type='daytime',
                activities=[
                  Activity(
                    description='Visit Golden Gate Bridge.'
                  ),
                  Activity(
                    description="Visit Fisherman's Wharf."
                  ),
                  Activity(
                    description='Visit Alcatraz Island.'
                  )
                ],
                hotel=None
              )
              ```

            NATURAL_LANGUAGE_TEXT:
            """),
    )

  def test_render_no_examples(self):
    value = Itinerary(
        day=1,
        type='daytime',
        activities=[
            Activity('Visit Golden Gate Bridge.'),
            Activity("Visit Fisherman's Wharf."),
            Activity('Visit Alcatraz Island.'),
        ],
        hotel=None,
    )
    l = description_lib._DescribeStructure(
        input=value, context='1 day itinerary to SF'
    )
    self.assertEqual(
        l.render().text,
        inspect.cleandoc("""
            Please help describe PYTHON_OBJECT in natural language.

            INSTRUCTIONS:
              1. Do not add details which are not present in the object.
              2. If a field in the object has None as its value, do not mention it.

            CONTEXT_FOR_DESCRIPTION:
              1 day itinerary to SF

            PYTHON_OBJECT:
              ```python
              Itinerary(
                day=1,
                type='daytime',
                activities=[
                  Activity(
                    description='Visit Golden Gate Bridge.'
                  ),
                  Activity(
                    description="Visit Fisherman's Wharf."
                  ),
                  Activity(
                    description='Visit Alcatraz Island.'
                  )
                ],
                hotel=None
              )
              ```

            NATURAL_LANGUAGE_TEXT:
            """),
    )

  def test_render_no_context(self):
    value = Itinerary(
        day=1,
        type='daytime',
        activities=[
            Activity('Visit Golden Gate Bridge.'),
            Activity("Visit Fisherman's Wharf."),
            Activity('Visit Alcatraz Island.'),
        ],
        hotel=None,
    )
    l = description_lib._DescribeStructure(input=value)
    self.assertEqual(
        l.render().text,
        inspect.cleandoc("""
            Please help describe PYTHON_OBJECT in natural language.

            INSTRUCTIONS:
              1. Do not add details which are not present in the object.
              2. If a field in the object has None as its value, do not mention it.

            PYTHON_OBJECT:
              ```python
              Itinerary(
                day=1,
                type='daytime',
                activities=[
                  Activity(
                    description='Visit Golden Gate Bridge.'
                  ),
                  Activity(
                    description="Visit Fisherman's Wharf."
                  ),
                  Activity(
                    description='Visit Alcatraz Island.'
                  )
                ],
                hotel=None
              )
              ```

            NATURAL_LANGUAGE_TEXT:
            """),
    )

  def test_describe(self):
    lm = fake.StaticSequence(
        [
            (
                "On the first day, visit the Golden Gate Bridge, Fisherman's"
                ' Wharf, and Alcatraz Island.'
            )
        ]
    )
    self.assertEqual(
        description_lib.describe(
            Itinerary(
                day=1,
                type='daytime',
                activities=[
                    Activity('Visit Golden Gate Bridge.'),
                    Activity("Visit Fisherman's Wharf."),
                    Activity('Visit Alcatraz Island.'),
                ],
                hotel=None,
            ),
            '1 day itinerary to SF',
            lm=lm,
        ),
        (
            "On the first day, visit the Golden Gate Bridge, Fisherman's Wharf,"
            ' and Alcatraz Island.'
        ),
    )


if __name__ == '__main__':
  unittest.main()



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Language models from DeepSeek."""

import datetime
import functools
import os
from typing import Annotated, Any, Final

import langfun.core as lf
from langfun.core.llms import openai_compatible
import pyglove as pg


class DeepSeekModelInfo(lf.ModelInfo):
  """DeepSeek model info."""

  LINKS = dict(
      models='https://api-docs.deepseek.com/quick_start/pricing',
      pricing='https://api-docs.deepseek.com/quick_start/pricing',
      rate_limits='https://api-docs.deepseek.com/quick_start/rate_limit',
      error_codes='https://api-docs.deepseek.com/quick_start/error_codes',
  )

  provider: Final[str] = 'DeepSeek'  # pylint: disable=invalid-name

  api_model_name: Annotated[
      str,
      'The model name used in the DeepSeek API.'
  ]


SUPPORTED_MODELS = [
    DeepSeekModelInfo(
        model_id='deepseek-r1',
        in_service=True,
        model_type='thinking',
        api_model_name='deepseek-reasoner',
        description='DeepSeek Reasoner model (01/20/2025).',
        url='https://api-docs.deepseek.com/news/news250120',
        release_date=datetime.datetime(2025, 1, 20),
        input_modalities=lf.ModelInfo.TEXT_INPUT_ONLY,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=64_000,
            max_output_tokens=8_000,
            max_cot_tokens=32_000,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=0.14,
            cost_per_1m_input_tokens=0.55,
            cost_per_1m_output_tokens=2.19,
        ),
        # No rate limits is enforced by DeepSeek for now.
        rate_limits=None
    ),
    DeepSeekModelInfo(
        model_id='deepseek-v3',
        in_service=True,
        model_type='instruction-tuned',
        api_model_name='deepseek-chat',
        description='DeepSeek V3 model (12/26/2024).',
        url='https://api-docs.deepseek.com/news/news1226',
        release_date=datetime.datetime(2024, 12, 26),
        input_modalities=lf.ModelInfo.TEXT_INPUT_ONLY,
        context_length=lf.ModelInfo.ContextLength(
            max_input_tokens=64_000,
            max_output_tokens=8_000,
        ),
        pricing=lf.ModelInfo.Pricing(
            cost_per_1m_cached_input_tokens=0.07,
            cost_per_1m_input_tokens=0.27,
            cost_per_1m_output_tokens=1.1,
        ),
        # No rate limits is enforced by DeepSeek for now.
        rate_limits=None
    ),
]

_SUPPORTED_MODELS_BY_ID = {m.model_id: m for m in SUPPORTED_MODELS}


# DeepSeek API uses an API format compatible with OpenAI.
# Reference: https://api-docs.deepseek.com/
@lf.use_init_args(['model'])
class DeepSeek(openai_compatible.OpenAICompatible):
  """DeepSeek model."""

  model: pg.typing.Annotated[
      pg.typing.Enum(
          pg.MISSING_VALUE, [m.model_id for m in SUPPORTED_MODELS]
      ),
      'The name of the model to use.',
  ]

  api_endpoint: str = 'https://api.deepseek.com/chat/completions'

  api_key: Annotated[
      str | None,
      (
          'API key. If None, the key will be read from environment variable '
          "'DEEPSEEK_API_KEY'."
      ),
  ] = None

  @property
  def headers(self) -> dict[str, Any]:
    api_key = self.api_key or os.environ.get('DEEPSEEK_API_KEY', None)
    if not api_key:
      raise ValueError(
          'Please specify `api_key` during `__init__` or set environment '
          'variable `DEEPSEEK_API_KEY` with your DeepSeek API key.'
      )
    headers = super().headers
    headers.update({
        'Authorization': f'Bearer {api_key}',
    })
    return headers

  @functools.cached_property
  def model_info(self) -> DeepSeekModelInfo:
    return _SUPPORTED_MODELS_BY_ID[self.model]

  def _request_args(
      self, options: lf.LMSamplingOptions) -> dict[str, Any]:
    """Returns a dict as request arguments."""
    # NOTE(daiyip): Replace model name with the API model name instead of the
    # model ID.
    args = super()._request_args(options)
    args['model'] = self.model_info.api_model_name
    return args

  @classmethod
  def dir(cls):
    return [m.model_id for m in SUPPORTED_MODELS if m.in_service]


class DeepSeekR1(DeepSeek):
  """DeepSeek Reasoner model.

  Currently it is powered by DeepSeek-R1 model, 64k input context, 8k max
  output, 32k max CoT output.
  """

  model = 'deepseek-r1'


class DeepSeekV3(DeepSeek):
  """DeepSeek Chat model.

  Currently, it is powered by DeepSeek-V3 model, 64K input contenxt window and
  8k max output tokens.
  """

  model = 'deepseek-v3'


def _register_deepseek_models():
  """Registers DeepSeek models."""
  for m in SUPPORTED_MODELS:
    lf.LanguageModel.register(m.model_id, DeepSeek)

_register_deepseek_models()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Video modality."""

import functools
from langfun.core.modalities import mime


class Video(mime.Mime):
  """Video."""

  MIME_PREFIX = 'video'

  @functools.cached_property
  def video_format(self) -> str:
    return self.mime_type.removeprefix(self.MIME_PREFIX + '/')

  def _mime_control_for(self, uri: str) -> str:
    return f'<video controls> <source src="{uri}"> </video>'



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Structures for location objects in an image."""

import random
import unittest
from langfun.assistant.capabilities.gui import location


class CoordinateTest(unittest.TestCase):

  def test_basics(self):
    pt = location.Coordinate(1, 2)
    self.assertEqual(pt.x, 1)
    self.assertEqual(pt.y, 2)
    self.assertEqual(pt.as_tuple(), (1, 2))

  def test_random(self):
    bound = location.BBox(0, 0, 10, 10)
    rand = random.Random(0)
    for _ in range(10):
      pt = location.Coordinate.random(bound, rand=rand)
      self.assertIsInstance(pt, location.Coordinate)
      self.assertIn(pt, bound)

  def test_from_value(self):
    pt = location.Coordinate.from_value((1, 2))
    self.assertEqual(pt.x, 1)
    self.assertEqual(pt.y, 2)
    self.assertIs(location.Coordinate.from_value(pt), pt)

  def test_arithmetic(self):
    pt1 = location.Coordinate(1, 2)
    pt2 = location.Coordinate(3, 4)
    self.assertEqual(pt1 + pt2, location.Coordinate(4, 6))
    self.assertEqual(pt2 + pt1, location.Coordinate(4, 6))
    self.assertEqual(pt1 - pt2, location.Coordinate(-2, -2))
    self.assertEqual(pt2 - pt1, location.Coordinate(2, 2))

    pt1 = location.Coordinate(1, 2)
    pt2 = (3, 4)
    self.assertEqual(pt1 + pt2, location.Coordinate(4, 6))
    self.assertEqual(pt2 + pt1, location.Coordinate(4, 6))
    self.assertEqual(pt1 - pt2, location.Coordinate(-2, -2))
    self.assertEqual(pt2 - pt1, location.Coordinate(2, 2))

    self.assertEqual(pt1 * 2, location.Coordinate(2, 4))
    self.assertEqual(2 * pt1, location.Coordinate(2, 4))

  def test_distance_to(self):
    pt1 = location.Coordinate(1, 2)
    pt2 = location.Coordinate(3, 4)
    # compare float values with a tolerance of 1e-6
    self.assertAlmostEqual(pt1.distance_to(pt2), 2.8284271247461903, delta=1e-6)


class BBoxTest(unittest.TestCase):

  def test_invalid_bbox_creation(self):
    with self.assertRaisesRegex(AssertionError, '.*'):
      location.BBox(100, 50, 100, 450)  # Zero width.
    with self.assertRaisesRegex(AssertionError, '.*'):
      location.BBox(100, 50, 300, 50)  # Zero height.
    with self.assertRaisesRegex(AssertionError, '.*'):
      location.BBox(100, 50, 50, 50)  # Zero width and height

  def test_basics(self):
    bbox = location.BBox(100, 50, 300, 450)
    self.assertEqual(bbox.x, 100)
    self.assertEqual(bbox.y, 50)
    self.assertEqual(bbox.right, 300)
    self.assertEqual(bbox.bottom, 450)
    self.assertEqual(bbox.left, 100)
    self.assertEqual(bbox.top, 50)
    self.assertEqual(bbox.width, 200)
    self.assertEqual(bbox.height, 400)

    self.assertEqual(bbox.center, location.Coordinate(200, 250))
    self.assertEqual(bbox.top_left, location.Coordinate(100, 50))
    self.assertEqual(bbox.bottom_right, location.Coordinate(300, 450))
    self.assertEqual(bbox.area, 80000)
    self.assertEqual(bbox.as_tuple(), (100, 50, 300, 450))

  def test_contains(self):
    bbox = location.BBox(100, 50, 300, 450)
    self.assertIn((120, 200), bbox)
    self.assertIn(location.Coordinate(120, 200), bbox)
    self.assertNotIn(location.Coordinate(80, 200), bbox)
    self.assertNotIn(location.Coordinate(120, 20), bbox)
    self.assertNotIn(location.Coordinate(320, 200), bbox)
    self.assertNotIn(location.Coordinate(120, 470), bbox)

    self.assertIn((100, 50, 300, 450), bbox)
    self.assertIn(location.BBox(100, 50, 300, 450), bbox)
    self.assertIn(location.BBox(120, 70, 260, 410), bbox)
    self.assertNotIn(location.BBox(60, 70, 260, 410), bbox)
    self.assertNotIn(location.BBox(120, 45, 260, 410), bbox)
    self.assertNotIn(location.BBox(120, 60, 310, 410), bbox)
    self.assertNotIn(location.BBox(120, 60, 260, 470), bbox)
    self.assertNotIn(location.BBox(0, 0, 400, 500), bbox)

    with self.assertRaisesRegex(ValueError, 'Invalid tuple size'):
      _ = (1, 2, 3) in bbox

    with self.assertRaisesRegex(ValueError, 'Invalid type'):
      _ = 'abc' in bbox  # pytype: disable=unsupported-operands

  def test_intersects(self):
    bbox = location.BBox(100, 50, 300, 450)
    self.assertTrue(bbox.intersects(location.BBox(100, 50, 300, 450)))
    self.assertTrue(bbox.intersects(location.BBox(120, 70, 260, 410)))
    self.assertTrue(bbox.intersects(location.BBox(60, 70, 260, 470)))
    self.assertFalse(bbox.intersects(location.BBox(0, 70, 90, 410)))
    self.assertFalse(bbox.intersects(location.BBox(60, 10, 400, 30)))

  def test_clip(self):
    # BBox within image bounds
    bbox = location.BBox(100, 50, 300, 450)
    clipped_bbox = bbox.clip((400, 500))
    self.assertEqual(clipped_bbox, bbox)

    # BBox exceeds right and bottom bounds
    bbox = location.BBox(100, 50, 500, 600)
    clipped_bbox = bbox.clip((400, 500))
    self.assertEqual(clipped_bbox, location.BBox(100, 50, 400, 500))

    # BBox exceeds left and top bounds
    bbox = location.BBox(-10, -20, 300, 450)
    clipped_bbox = bbox.clip((400, 500))
    self.assertEqual(clipped_bbox, location.BBox(0, 0, 300, 450))

    # BBox larger than the image
    bbox = location.BBox(-10, -20, 800, 700)
    clipped_bbox = bbox.clip((400, 500))
    self.assertEqual(clipped_bbox, location.BBox(0, 0, 400, 500))

    # BBox starts outside the image boundaries (bottom/right)
    bbox = location.BBox(500, 600, 700, 800)
    clipped_bbox = bbox.clip((400, 500))
    self.assertIsNone(clipped_bbox)

  def test_random(self):
    bound = location.BBox(0, 0, 800, 600)
    rand = random.Random(0)
    for _ in range(10):
      bbox = location.BBox.random(bound, rand=rand)
      self.assertIsInstance(bbox, location.BBox)
      self.assertIn(bbox, bound)

    with self.assertRaisesRegex(
        ValueError, 'Minimum width or height is larger than the bound'
    ):
      _ = location.BBox.random(location.BBox(0, 0, 100, 100), min_width=110)

  def test_matches(self):
    bbox1 = location.BBox(100, 50, 300, 450)
    self.assertFalse(bbox1.matches(location.BBox(10, 10, 11, 11)))

    bbox2 = location.BBox(100, 50, 300, 450)
    self.assertTrue(bbox1.matches(bbox2))
    self.assertTrue(bbox2.matches(bbox1))

    bbox3 = location.BBox(110, 40, 290, 450)
    self.assertTrue(bbox1.matches(bbox3))
    self.assertFalse(bbox3.matches(bbox1, area_diff_threshold=0.05))
    self.assertFalse(bbox3.matches(bbox1, max_center_distance=1))

  def test_expand(self):
    # Test case 1: Expanding both width and height.
    bbox = location.BBox(100, 50, 300, 450)
    expanded_bbox = bbox.expand(width_scale=1.2, height_scale=1.1)
    self.assertEqual(expanded_bbox.x, 80)
    self.assertEqual(expanded_bbox.y, 30)
    self.assertEqual(expanded_bbox.right, 320)
    self.assertEqual(expanded_bbox.bottom, 470)
    self.assertEqual(expanded_bbox.width, 240)
    self.assertEqual(expanded_bbox.height, 440)
    self.assertEqual(expanded_bbox.center, location.Coordinate(200, 250))

    # Test case 2: Expanding only width.
    bbox = location.BBox(100, 50, 300, 450)
    expanded_bbox = bbox.expand(width_scale=1.2)
    self.assertEqual(expanded_bbox.x, 80)
    self.assertEqual(expanded_bbox.y, 50)
    self.assertEqual(expanded_bbox.right, 320)
    self.assertEqual(expanded_bbox.bottom, 450)
    self.assertEqual(expanded_bbox.width, 240)
    self.assertEqual(expanded_bbox.height, 400)
    self.assertEqual(expanded_bbox.center, location.Coordinate(200, 250))

    # Test case 3: Expanding only height.
    bbox = location.BBox(100, 50, 300, 450)
    expanded_bbox = bbox.expand(height_scale=1.1)
    self.assertEqual(expanded_bbox.x, 100)
    self.assertEqual(expanded_bbox.y, 30)
    self.assertEqual(expanded_bbox.right, 300)
    self.assertEqual(expanded_bbox.bottom, 470)
    self.assertEqual(expanded_bbox.width, 200)
    self.assertEqual(expanded_bbox.height, 440)
    self.assertEqual(expanded_bbox.center, location.Coordinate(200, 250))

    # Test case 4: Shrinking.
    bbox = location.BBox(100, 50, 300, 450)
    expanded_bbox = bbox.expand(width_scale=0.8, height_scale=0.9)
    self.assertEqual(expanded_bbox.x, 120)
    self.assertEqual(expanded_bbox.y, 70)
    self.assertEqual(expanded_bbox.right, 280)
    self.assertEqual(expanded_bbox.bottom, 430)
    self.assertEqual(expanded_bbox.width, 160)
    self.assertEqual(expanded_bbox.height, 360)
    self.assertEqual(expanded_bbox.center, location.Coordinate(200, 250))

    # Test case 5: No change.
    bbox = location.BBox(100, 50, 300, 450)
    expanded_bbox = bbox.expand(width_scale=1.0, height_scale=1.0)
    self.assertEqual(expanded_bbox, bbox)

if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Modalities for LLM."""

# pylint: disable=g-importing-member
# pylint: disable=g-bad-import-order
# pylint: disable=g-import-not-at-top

from langfun.core.modalities.audio import Audio
from langfun.core.modalities.mime import Mime
from langfun.core.modalities.mime import Custom
from langfun.core.modalities.image import Image
from langfun.core.modalities.pdf import PDF
from langfun.core.modalities.video import Video

# pylint: enable=g-import-not-at-top
# pylint: enable=g-bad-import-order
# pylint: enable=g-importing-member



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility library for handling concurrency in langfun."""

import abc
import collections
from collections.abc import Mapping
import concurrent.futures
import dataclasses
import io
import random
import sys
import threading
import time
from typing import Annotated, Any, Callable, Iterable, Iterator, Literal, Sequence, Tuple, Type, Union

import pyglove as pg


progress_bar: Literal['tqdm', 'console', None] = None

try:
  from tqdm import auto as tqdm   # pylint: disable=g-import-not-at-top
  progress_bar = 'tqdm'
except ImportError:
  progress_bar = 'console'
  tqdm = None


class RetryError(RuntimeError):
  """Retry error."""

  def __init__(
      self,
      func: Callable[..., Any],
      errors: list[BaseException],
      wait_intervals: list[int],
  ):
    assert len(errors) == len(wait_intervals) + 1

    super().__init__()
    self.func = func
    self.errors = errors
    self.wait_intervals = wait_intervals

  @property
  def attempts(self) -> int:
    """Returns the number of attempts made."""
    return len(self.errors)

  def __repr__(self) -> str:
    return (
        'RetryError('
        + pg.object_utils.kvlist_str(
            [
                ('func', self.func, None),
                ('errors', self.errors, None),
                ('wait_intervals', self.wait_intervals, None),
            ],
            compact=True,
        )
        + ')'
    )

  def __str__(self) -> str:
    wait_interval_str = ', '.join([str(x) for x in self.wait_intervals])
    return (
        f'Calling {self.func!r} failed after {self.attempts} attempts '
        f'(wait time: {wait_interval_str} seconds). '
        f'Last error: {self.errors[-1]}'
    )

  def __eq__(self, other: 'RetryError') -> bool:
    if not isinstance(other, RetryError):
      return False
    return (self.func is other.func
            and self.errors == other.errors
            and self.wait_intervals == other.wait_intervals)

  def __ne__(self, other: 'RetryError') -> bool:
    return not self.__eq__(other)

  def __hash__(self) -> int:
    return hash((
        RetryError, self.func, tuple(self.errors), tuple(self.wait_intervals)))


def with_retry(
    func: Callable[[Any], Any],
    retry_on_errors: Union[
        Union[Type[BaseException], Tuple[Type[BaseException], str]],
        Sequence[Union[Type[BaseException], Tuple[Type[BaseException], str]]],
    ],
    max_attempts: int,
    retry_interval: int | tuple[int, int] = (5, 60),
    exponential_backoff: bool = True,
    max_retry_interval: int = 300,
    seed: int | None = None,
) -> Callable[..., Any]:
  """Derives a user function with retry on error.

  Args:
    func: A user function.
    retry_on_errors: A sequence of exception types or tuples of exception type
      and error messages (described in regular expression) as the desired
      exception types to retry.
    max_attempts: Max number of attempts if an error to retry is encountered.
    retry_interval: The (base) retry interval in seconds. If a tuple, the retry
      interval will be randomly chosen between the first and the second element
      of the tuple.
    exponential_backoff: If True, exponential wait time will be applied on top
      of the base retry interval.
    max_retry_interval: The max retry interval in seconds. This is useful when
      the retry interval is exponential, to avoid the wait time to grow
      exponentially.
    seed: Random seed to generate retry interval. If None, the seed will be
      determined based on current time.

  Returns:
    A function with the same signature of the input function, with the retry
    capability.
  """

  def _func(*args, **kwargs):
    job = Job(
        func,
        args,
        kwargs,
        retry_on_errors=retry_on_errors,
        max_attempts=max_attempts,
        retry_interval=retry_interval,
        exponential_backoff=exponential_backoff,
        max_retry_interval=max_retry_interval,
        seed=seed,
    )
    job()
    if job.error:
      raise job.error
    return job.result

  return _func


class RetryEntry(pg.Object):
  """Retry entry."""

  call_interval: float
  error: BaseException | None = None
  wait_interval: float = 0.


def concurrent_execute(
    func: Callable[[Any], Any],
    parallel_inputs: Iterable[Any],
    *,
    executor: Union[concurrent.futures.ThreadPoolExecutor, str, None] = None,
    max_workers: int = 32,
    retry_on_errors: Union[
        Union[Type[BaseException], Tuple[Type[BaseException], str]],
        Sequence[Union[Type[BaseException], Tuple[Type[BaseException], str]]],
        None,
    ] = None,
    max_attempts: int = 5,
    retry_interval: int | tuple[int, int] = (5, 60),
    exponential_backoff: bool = True,
    max_retry_interval: int = 300,
    return_jobs: bool = False,
) -> list[Any]:
  """Executes a function concurrently under current component context.

  Args:
    func: A user function.
    parallel_inputs: The inputs for `func` which will be processed in parallel.
    executor: A thread pool executor or a resource ID string to pool work items.
      When resource ID is used, a thread pool will be created and cached in a
      executor pool for future reuse. If None, a new thread pool executor will
      be created for current execution.
    max_workers: The max number of workers.
    retry_on_errors: A sequence of exception types or tuples of exception type
      and error messages (described in regular expression) as the desired
      exception types to retry.
    max_attempts: Max number of attempts if an error to retry is encountered.
    retry_interval: The (base) retry interval in seconds. If a tuple, the retry
      interval will be randomly chosen between the first and the second element
      of the tuple.
    exponential_backoff: If True, exponential wait time will be applied on top
      of the base retry interval.
    max_retry_interval: The max retry interval in seconds. This is useful when
      the retry interval is exponential, to avoid the wait time to grow
      exponentially.
    return_jobs: If True, return a list of `Job` objects. Otherwise, return a
      list of outputs.

  Returns:
    A list of ouputs. Each is the return value of `func` based on the input
      value. Order is preserved.
  """
  jobs = []
  for inputs in parallel_inputs:
    jobs.append(
        Job(
            func,
            (inputs,),
            retry_on_errors=retry_on_errors,
            max_attempts=max_attempts,
            retry_interval=retry_interval,
            exponential_backoff=exponential_backoff,
            max_retry_interval=max_retry_interval,
        )
    )

  # NOTE(daiyip): when executor is not specified and max_worker is 1,
  # we don't need to create a executor pool. Instead, the inputs will be
  # processed by the user function in sequence within the current thread.
  if executor is None and max_workers == 1:
    for job in jobs:
      job()
      if job.error:
        raise job.error
    return jobs if return_jobs else [job.result for job in jobs]

  shutdown_after_finish = executor is None
  executor = _executor_pool.executor_from(executor, max_workers=max_workers)

  try:
    executed_jobs = list(
        executor.map(
            lambda job: job(),
            [pg.with_contextual_override(job) for job in jobs]
        )
    )
    for job in executed_jobs:
      if job.error:
        raise job.error
    return (
        executed_jobs if return_jobs else [job.result for job in executed_jobs]
    )
  finally:
    if shutdown_after_finish:
      # Do not wait threads to finish if they are timed out.
      executor.shutdown(wait=False, cancel_futures=True)


@dataclasses.dataclass
class Job:
  """Thread pool job."""

  func: Callable[[Any], Any]
  args: Sequence[Any] = ()
  kwargs: Mapping[str, Any] = dataclasses.field(default_factory=dict)
  _: dataclasses.KW_ONLY

  result: Any = pg.MISSING_VALUE
  error: Annotated[
      BaseException | None,
      'The non-retryable error encountered during the job execution.',
  ] = None
  retry_entries: Annotated[
      Sequence[RetryEntry], 'Records of retry attempts.'
  ] = dataclasses.field(default_factory=list)

  retry_on_errors: Annotated[
      Sequence[Type[BaseException] | str],
      (
          'A sequence of exception types or tuples of exception type and error '
          'messages (described in regular expression) as the desired exception '
          'types to retry.'
      ),
  ] = ()
  max_attempts: Annotated[
      int, 'Max number of attempts if an error to retry is encountered.'
  ] = 5
  retry_interval: Annotated[
      int | tuple[int, int],
      (
          'The (base) retry interval in seconds. If a tuple, the retry '
          'interval will be randomly chosen between the first and the second '
          'element of the tuple.'
      ),
  ] = (5, 60)
  exponential_backoff: Annotated[
      bool,
      (
          'If True, exponential wait time will be applied on top of the base '
          'retry interval.'
      ),
  ] = True
  max_retry_interval: Annotated[
      int,
      (
          'The max retry interval in seconds. This is useful when the retry '
          'interval is exponential, to avoid the wait time to grow '
          'exponentially.'
      ),
  ] = 300
  seed: Annotated[
      int | None,
      (
          'Random seed to generate retry interval. If None, the seed will be'
          ' determined based on current time.'
      ),
  ] = None

  timeit: pg.object_utils.TimeIt = dataclasses.field(
      default_factory=lambda: pg.object_utils.TimeIt('job')
  )

  @property
  def elapse(self) -> float:
    """Returns the running time in seconds since the job get started."""
    return self.timeit.elapse

  def _retry_call(self) -> 'Job':
    """Retries func call on args."""
    rand = random if self.seed is None else random.Random(self.seed)

    def base_interval() -> int:
      if isinstance(self.retry_interval, tuple):
        return rand.randint(*self.retry_interval)
      else:
        assert isinstance(self.retry_interval, int)
        return self.retry_interval

    def next_wait_interval(attempt: int) -> float:
      if not self.exponential_backoff:
        attempt = 1
      return min(
          self.max_retry_interval, base_interval() * (2 ** (attempt - 1))
      )

    retry_entries = []
    wait_interval = 0
    while True:
      with pg.catch_errors(self.retry_on_errors) as error_context:
        begin_time = time.time()
        self.result = self.func(*self.args, **self.kwargs)

      end_time = time.time()
      retry_entries.append(RetryEntry(
          call_interval=end_time - begin_time,
          wait_interval=wait_interval,
          error=error_context.error,
      ))
      if error_context.error is None:
        self.retry_entries = retry_entries
        return self

      # Branch when errors are met for retry.
      if len(retry_entries) < self.max_attempts:
        wait_interval = next_wait_interval(len(retry_entries))

        pg.logging.warning(
            f'Calling {self.func!r} encountered {error_context.error!r} '
            f'(attempts={len(retry_entries)}), retrying in '
            f'{wait_interval} seconds...'
        )

        time.sleep(wait_interval)
      else:
        errors = [e.error for e in retry_entries]
        # First wait interval is 0.
        wait_intervals = [e.wait_interval for e in retry_entries[1:]]
        raise RetryError(self.func, errors, wait_intervals)

  def __call__(self) -> 'Job':
    if getattr(self, '_has_call', False):
      raise ValueError('Job can only be called once.')
    self._has_call = True
    try:
      with self.timeit:
        if self.retry_on_errors:
          return self._retry_call()
        self.result = self.func(*self.args, **self.kwargs)
    except BaseException as e:  # pylint: disable=broad-exception-caught
      self.error = e
    return self

  def mark_canceled(self, error: BaseException) -> None:
    """Marks the job as canceled."""
    self.timeit.end(error)
    self.error = error


@dataclasses.dataclass
class Progress:
  """Concurrent processing progress."""
  total: int

  _succeeded: int = 0
  _failed: int = 0
  _last_error: BaseException | None = None
  _total_duration: float = 0.0
  _job: Job | None = None
  _timeit_summary: pg.object_utils.TimeIt.StatusSummary = dataclasses.field(
      default_factory=pg.object_utils.TimeIt.StatusSummary
  )

  @property
  def succeeded(self) -> int:
    """Returns number of succeeded jobs."""
    return self._succeeded

  @property
  def failed(self) -> int:
    """Returns number of failed jobs."""
    return self._failed

  @property
  def completed(self) -> int:
    """Returns number of completed jobs."""
    return self.succeeded + self.failed

  @property
  def last_error(self) -> BaseException | None:
    """Returns last error."""
    return self._last_error

  @property
  def job(self) -> Job | None:
    """Returns current job."""
    return self._job

  @property
  def success_rate(self) -> float:
    """Returns success rate."""
    if self.completed == 0:
      return 0.0
    return self.succeeded / self.completed

  @property
  def failure_rate(self) -> float:
    """Returns failure rate."""
    if self.completed == 0:
      return 0.0
    return self.failed / self.completed

  @property
  def avg_duration(self) -> float:
    """Returns average duration each job worked."""
    if self.completed == 0:
      return 0.0
    return self._total_duration / self.completed

  @property
  def timeit_summary(self) -> pg.object_utils.TimeIt.StatusSummary:
    """Returns the aggregated summary for each `pg.timeit`."""
    return self._timeit_summary

  def timeit_summary_str(self) -> str | None:
    if not self.timeit_summary:
      return None
    return ', '.join([
        '%s (%.2fs, %d/%d)' % (
            k.lstrip('job.'), v.avg_duration, v.num_ended, v.num_started
        ) for k, v in self.timeit_summary.breakdown.items() if k != 'job'
    ])

  def last_error_str(self) -> str | None:
    if self.last_error is None:
      return None
    error_text = repr(self.last_error)
    if len(error_text) >= 64:
      error_text = error_text[:64] + '...'
    return error_text

  def update(self, job: Job) -> None:
    """Mark a job as completed."""
    self._job = job
    if job.error is None:
      self._succeeded += 1
    else:
      self._failed += 1
      self._last_error = job.error
    self._total_duration += job.elapse
    self._timeit_summary.aggregate(job.timeit.status())


class ProgressBar:
  """Progress bars that can be updated in concurrent threads.

  When progresses are reported during `lf.concurrent_map`, thread functions may
  have child calls to `lf.concurrent_map`, which requires multiple progress bar
  created/updated in multi-threads to be supported.

  However, tqdm requires all progress bars to be created and updated in the
  main thread. This class uses a queue system to communicate among the threads,
  allowing bar installation/uninstallation/update to be called within non-main
  thread.
  """

  @dataclasses.dataclass
  class Settings:
    """Progress bar settings."""
    label: str | None
    total: int
    color: str | None = None
    status: dict[str, Any] | None = None

  @dataclasses.dataclass
  class Update:
    """Progress bar update."""
    bar_id: int
    delta: int
    status: Union[dict[str, Any], str, None] = None
    color: str | None = None

  _progress_bars: dict[int, '_ProgressControl'] = {}
  _install_requests: list[tuple[int, Settings]] = []
  _updates: collections.deque[Update] = collections.deque()
  _uninstall_requests: list[int] = []
  _lock = threading.Lock()

  @classmethod
  def install(
      cls,
      label: str | None,
      total: int,
      color: str | None = None,
      status: dict[str, Any] | None = None,
      ) -> int:
    """Installs a progress bar and returns a reference id."""
    with cls._lock:
      settings = ProgressBar.Settings(label, total, color, status)
      bar_id = id(settings)
      cls._install_requests.append((bar_id, settings))
      return bar_id

  @classmethod
  def update(
      cls,
      bar_id: int,
      delta: int = 0,
      status: Union[dict[str, Any], str, None] = None,
      color: str | None = None,
      refresh: bool = True,
      ) -> None:
    """Report the progress for a label."""
    if status is not None and not isinstance(status, (str, dict)):
      raise ValueError(f'Unsupported status: {status}')
    with cls._lock:
      cls._updates.append(
          ProgressBar.Update(
              bar_id=bar_id, delta=delta, status=status, color=color,
          )
      )
    if refresh:
      cls.refresh()

  @classmethod
  def uninstall(cls, bar_id: int) -> None:
    """Uninstalls a progress bar with the reference id returned by install."""
    with cls._lock:
      cls._uninstall_requests.append(bar_id)

  @classmethod
  def refresh(cls) -> None:
    """Update all progress bars when called within the main thread."""
    if threading.current_thread() is not threading.main_thread():
      return

    with cls._lock:
      # Process install requests.
      if cls._install_requests:
        for bar_id, settings in cls._install_requests:
          cls._progress_bars[bar_id] = _progress_control(
              total=settings.total,
              label=settings.label,
              color=settings.color,
              status=settings.status)
        cls._install_requests.clear()

      # Process updates.
      updated_bars = set()
      while cls._updates:
        update = cls._updates.popleft()
        bar = cls._progress_bars.get(update.bar_id)
        # Processing of updates may be delayed, in such case the bar might
        # be already uninstalled from a different thread.
        if bar is None:
          continue
        if update.delta > 0:
          bar.update(update.delta)

        if update.status is not None:
          bar.set_status(update.status)

        if update.color is not None:
          bar.set_color(update.color)
        updated_bars.add(bar)

      # Refresh each updated bar just once.
      for bar in updated_bars:
        bar.refresh()

      # Process uninstall requests.
      if cls._uninstall_requests:
        for bar_id in cls._uninstall_requests:
          bar = cls._progress_bars.pop(bar_id, None)
          if bar is not None:
            bar.close()
        cls._uninstall_requests.clear()


def concurrent_map(
    func: Callable[[Any], Any],
    parallel_inputs: Iterable[Any],
    *,
    executor: Union[concurrent.futures.ThreadPoolExecutor, str, None] = None,
    max_workers: int = 32,
    ordered: bool = False,
    show_progress: bool | int = False,
    label: str | None = None,
    color: Literal[
        'red',
        'blue',
        'green',
        'black',
        'yellow',
        'magenta',
        'cyan',
        'white',
        None,
    ] = None,
    status_fn: Callable[[Progress], dict[str, Any]] | None = None,
    timeout: int | None = None,
    silence_on_errors: Union[
        Type[BaseException], Tuple[Type[BaseException], ...], None
    ] = Exception,
    retry_on_errors: Union[
        Type[BaseException],
        Tuple[Type[BaseException], ...],
        None,
    ] = None,
    max_attempts: int = 5,
    retry_interval: int | tuple[int, int] = (5, 60),
    exponential_backoff: bool = True,
    return_jobs: bool = False,
) -> Iterator[Any]:
  """Maps inputs to outptus via func concurrently under current context.

  Args:
    func: A user function.
    parallel_inputs: The inputs for `func` which will be processed in parallel.
    executor: A thread pool executor or a resource ID string to pool work items.
      When resource ID is used, a thread pool will be created and cached in a
      executor pool for future reuse. If None, a new thread pool executor will
      be created for current execution.
    max_workers: The max number of workers.
    ordered: If True, the returned iterator will emit (input, output, error) in
      the order of the elements in `parallel_inputs`. Otherwise, elements that
      are finished earlier will be delivered first.
    show_progress: A boolean or an integer as the bar id returned from
      `lf.concurrent.ProgressBar.install`. If True, a bar will be created on the
      fly showing the progress of the execution. If False, no progress bar will
      be shown. If a bar id, the progress update will be associated with the
      bar.
    label: An optional label for the progress bar. Applicable when
      `show_progress` is set to True.
    color: Color of the progress bar. Applicable when `show_progress` is set to
      True or a bar id.
    status_fn: An optional callable object that receives a
      `lf.concurrent.Progress` object and returns a dict of kv pairs as the
      status to include in the progress bar. Applicable only when
      `show_progress` is set to True or a bar id. If None, the default status_fn
      will be used, which outputs the success and failure rate.
    timeout: The timeout in seconds for processing each input. It is the total
      processing time for each input, even multiple retries take place. If None,
      there is no timeout.
    silence_on_errors: If None, any errors raised during processing any inputs
      will be raised. Otherwise, the matched errors will not raise, instead,
      they will be returned as the third_element of the tuple.
    retry_on_errors: A sequence of exception types or tuples of exception type
      and error messages (described in regular expression) as the desired
      exception types to retry. These errors are usually transient error.
    max_attempts: Max number of attempts if an error to retry is encountered.
    retry_interval: The (base) retry interval in seconds. If a tuple, the retry
      interval will be randomly chosen between the first and the second element
      of the tuple.
    exponential_backoff: If True, exponential wait time will be applied on top
      of the base retry interval.
    return_jobs: If True, the returned iterator will emit `Job` objects.

  Yields:
    An iterator of (input, output, error) or Job object.

  Raises:
    Exception: Errors that are not in `silence_on_errors` or `retry_on_errors`,
      or retry on such errors has reached max attempts.
    TimeoutError: Any item timed out while TimeoutError is not silenced via
      `silence_on_errors`.
  """
    # Internal usage logging.

  status_fn = status_fn or (lambda p: {   # pylint: disable=g-long-lambda
      'Succeeded': '%.2f%% (%d/%d)' % (
          p.success_rate * 100, p.succeeded, p.completed),
      'Failed': '%.2f%% (%d/%d)' % (
          p.failure_rate * 100, p.failed, p.completed),
  })

  shutdown_after_finish = executor is None
  executor = _executor_pool.executor_from(executor, max_workers=max_workers)

  future_to_job = {}
  pending_futures = []
  total = 0
  for inputs in parallel_inputs:
    job = Job(
        func,
        (inputs,),
        retry_on_errors=retry_on_errors,
        max_attempts=max_attempts,
        retry_interval=retry_interval,
        exponential_backoff=exponential_backoff,
    )
    future = executor.submit(pg.with_contextual_override(job))
    pending_futures.append(future)
    future_to_job[future] = job
    total += 1

  # Setup progress bar.
  progress = Progress(total=total)
  if isinstance(show_progress, bool):
    external_bar = False
    bar_id = ProgressBar.install(label, total, color) if show_progress else None
  else:
    bar_id = show_progress
    external_bar = True
    show_progress = True

  def update_progress_bar(progress: Progress) -> None:
    if show_progress:
      status = status_fn(progress)
      status.update({
          'AvgDuration': '%.2fs' % progress.avg_duration
      })
      if progress.last_error is not None:
        status['LastError'] = progress.last_error_str()

      if progress.timeit_summary:
        status['TimeIt'] = progress.timeit_summary_str()
      ProgressBar.update(bar_id, delta=1, status=status)

  try:
    if ordered:
      for future in pending_futures:
        job = future_to_job[future]
        completed = False
        while True:
          try:
            _ = future.result(timeout=1)
            completed = True
          except concurrent.futures.TimeoutError:
            if timeout and timeout < job.elapse:
              future.cancel()
              last_error = TimeoutError(
                  f'Execution time ({job.elapse}) exceeds {timeout} seconds.')
              job.mark_canceled(last_error)
              completed = True

          if completed:
            if job.error is not None and not (
                silence_on_errors and isinstance(job.error, silence_on_errors)):
              raise job.error   # pylint: disable=g-doc-exception

            yield job if return_jobs else job.args[0], job.result, job.error
            progress.update(job)
            update_progress_bar(progress)
            ProgressBar.refresh()
            break
          else:
            # There might be updates from other concurrent_map. So even there
            # is no progress on current map, we still update the progress
            # manager.
            ProgressBar.refresh()
    else:
      while pending_futures:
        completed_batch = set()
        try:
          for future in concurrent.futures.as_completed(
              pending_futures, timeout=1):
            job = future_to_job[future]
            del future_to_job[future]
            if job.error is not None and not (
                silence_on_errors and isinstance(job.error, silence_on_errors)):
              raise job.error   # pylint: disable=g-doc-exception
            yield job if return_jobs else job.args[0], job.result, job.error
            progress.update(job)
            update_progress_bar(progress)
            completed_batch.add(future)
            ProgressBar.refresh()
        except concurrent.futures.TimeoutError:
          pass

        remaining_futures = []
        for future in pending_futures:
          if future in completed_batch:
            continue
          job = future_to_job[future]
          if timeout and job.elapse > timeout:
            if not future.done():
              future.cancel()
              job.mark_canceled(
                  TimeoutError(f'Execution time ({job.elapse}) '
                               f'exceeds {timeout} seconds.'))
              if not (silence_on_errors
                      and isinstance(job.error, silence_on_errors)):
                raise job.error  # pylint: disable=g-doc-exception

            yield job.args[0], job.result, job.error
            progress.update(job)
            update_progress_bar(progress)
          else:
            remaining_futures.append(future)
        pending_futures = remaining_futures
        ProgressBar.refresh()
  finally:
    if show_progress and not external_bar:
      ProgressBar.uninstall(bar_id)

    if shutdown_after_finish:
      # Do not wait threads to finish if they are timed out.
      executor.shutdown(wait=False, cancel_futures=True)


class ExecutorPool:
  """A pool of managed executors.

  Managed executors are used for controlling the parallelism of execution based
  on resource id. This design is to honor overall rate limit of LMs globally
  (with current process).
  """

  def __init__(self):
    self._executors: dict[str, concurrent.futures.ThreadPoolExecutor] = {}

  def get(
      self, resource_id: str, max_workers: int | None = None
  ) -> concurrent.futures.ThreadPoolExecutor:
    """Gets or creates a thread pool executor associated with a resource id."""
    executor = self._executors.get(resource_id)
    if executor is None:
      executor = concurrent.futures.ThreadPoolExecutor(max_workers=max_workers)
      self._executors[resource_id] = executor
    return executor

  @property
  def resource_ids(self) -> list[str]:
    """Returns the resource ids for active executors."""
    return list(self._executors.keys())

  def executor_from(
      self,
      maybe_executor: Union[concurrent.futures.ThreadPoolExecutor, str, None],
      max_workers: int | None = None,
  ) -> concurrent.futures.ThreadPoolExecutor:
    """Creates a thread pool executor."""
    if isinstance(maybe_executor, concurrent.futures.ThreadPoolExecutor):
      return maybe_executor
    elif isinstance(maybe_executor, str):
      return self.get(maybe_executor, max_workers)
    elif maybe_executor is None:
      return concurrent.futures.ThreadPoolExecutor(max_workers=max_workers)
    else:
      raise ValueError(f'Unsupported value: {maybe_executor}.')


class _ProgressControl(pg.Object):
  """Abstract progress control."""
  # Disable symbolic comparison so the hash is based on object address.
  use_symbolic_comparison = False

  total: int
  label: str | None
  color: str | None
  status: str | dict[str, Any] | None

  def set_color(self, color: str | None):
    with pg.notify_on_change(False):
      self.rebind(color=color)

  def set_status(self, status: str | dict[str, Any] | None):
    with pg.notify_on_change(False):
      self.rebind(status=status)

  @abc.abstractmethod
  def update(self, delta):
    """Update progress."""

  @abc.abstractmethod
  def refresh(self) -> None:
    """Refresh progress bar."""

  @abc.abstractmethod
  def close(self) -> None:
    """Close progress bar."""


class _TqdmProgressControl(_ProgressControl):
  """Tqdm-based progress control."""

  def _on_bound(self):
    super()._on_bound()
    assert tqdm is not None
    self._tqdm = tqdm.tqdm(
        total=self.total,
        desc=self.label,
        colour=self.color,
        postfix=self.status,
    )

  def update(self, delta: int) -> None:
    self._tqdm.update(delta)

  def refresh(self):
    self._tqdm.set_description(self.label, refresh=False)
    if isinstance(self.status, str):
      self._tqdm.set_postfix_str(self.status, refresh=False)
    else:
      self._tqdm.set_postfix(self.status, refresh=False)
    self._tqdm.colour = self.color
    self._tqdm.refresh()

  def close(self):
    self._tqdm.close()


class _ConsoleProgressControl(_ProgressControl):
  """Simple progress control by printing the status to the console."""

  def _on_bound(self):
    super()._on_bound()
    self._progress = 0

  def update(self, delta: int) -> None:
    self._progress += delta

  def refresh(self):
    s = io.StringIO()
    if self.label is not None:
      s.write(pg.colored(self.label, 'red', styles=['bold']))
      s.write(': ')
    s.write(
        pg.colored(
            '%d%% (%d/%d)' %
            (
                self._progress * 100 // self.total,
                self._progress,
                self.total,
            ),
            color=self.color or 'green'
        )
    )
    if self.status is not None:
      status = repr(self.status) if isinstance(
          self.status, dict) else self.status
      s.write(f' : {status}')
    sys.stderr.write(s.getvalue() + '\n')

  def close(self):
    sys.stderr.flush()


class _NoopProgressControl(_ProgressControl):
  """No-op progress control."""

  def update(self, delta: int) -> None:
    pass

  def refresh(self) -> None:
    pass

  def close(self) -> None:
    pass


def _progress_control(
    total: int,
    label: str | None,
    color: str | None,
    status: str | dict[str, Any] | None,
) -> _ProgressControl:
  """Creates a process control."""
  if progress_bar == 'tqdm':
    if not tqdm:
      raise RuntimeError(
          'Please install package "tqdm" to use `tqdm` progress bar.'
      )
    return _TqdmProgressControl(total, label, color, status)
  elif progress_bar == 'console':
    return _ConsoleProgressControl(total, label, color, status)
  elif progress_bar is None:
    return _NoopProgressControl(total, label, color, status)
  else:
    raise ValueError(f'Unsupported progress bar type: {progress_bar}')


def get_executor(
    resource_id: str,
    max_workers: int | None = None) -> concurrent.futures.ThreadPoolExecutor:
  """Gets a thread pool executor associated with a resource id."""
  return _executor_pool.get(resource_id, max_workers)

# The global executor pool based on resource IDs.
_executor_pool = ExecutorPool()
